import pandas as pd

from gtime.feature_extraction import Shift


def horizon_shift(time_series: pd.DataFrame, horizon: int = 5) -> pd.DataFrame:
    """Perform a shift of the original ``time_series`` for each time step between 1 and
    ``horizon``.

    Parameters
    ----------
    time_series : pd.DataFrame, shape (n_samples, n_features), required
        The list of ``TimeSeriesFeature`` from which to compute the feature_extraction.

    horizon : int, optional, default: ``5``
        It represents how much into the future is necessary to predict. This corresponds
        to the number of shifts that are going to be performed on y.
        
    Returns
    -------
    y : pd.DataFrame, shape (n_samples, horizon)
        The shifted time series.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.model_selection import horizon_shift
    >>> X = pd.DataFrame(range(0, 5), index=pd.date_range("2020-01-01", "2020-01-05"))
    >>> horizon_shift(X, horizon=2)
                y_1  y_2
    2020-01-01  1.0  2.0
    2020-01-02  2.0  3.0
    2020-01-03  3.0  4.0
    2020-01-04  4.0  NaN
    2020-01-05  NaN  NaN

    """
    y = pd.DataFrame(index=time_series.index)
    for k in range(1, horizon + 1):
        shift_feature = Shift(-k)
        y[f"y_{k}"] = shift_feature.fit_transform(time_series)

    return y



import numpy as np
import pandas as pd
if pd.__version__ >= '1.0.0':
    import pandas._testing as testing
else:
    import pandas.util.testing as testing


from gtime.forecasting import TrendForecaster


def test_polynomial_trend():
    testing.N, testing.K = 500, 1
    df = testing.makeTimeDataFrame(freq="D")

    df["A"] = df["A"] + 0.0005 * pd.Series(
        index=df.index, data=range(df.shape[0])
    ) * pd.Series(index=df.index, data=range(df.shape[0]))

    tm = TrendForecaster(trend="polynomial", trend_x0=np.zeros(5))
    tm.fit(df)

    assert np.allclose(tm.best_trend_params_, [0.0] * len(tm.best_trend_params_))


def test_exponential_trend():
    testing.N, testing.K = 500, 1
    df = testing.makeTimeDataFrame(freq="D")

    df["A"] = df["A"] + 0.0005 * pd.Series(
        index=df.index, data=range(df.shape[0])
    ).apply(lambda x: np.exp(0.03 * x))

    tm = TrendForecaster(trend="exponential", trend_x0=0)
    tm.fit(df)

    assert np.allclose(tm.best_trend_params_, 0.0)


# TODO: predicting tests



import numpy as np


def polynomial(X, weights):
    return np.poly1d(weights)(X)


def exponential(X, exponent):
    return np.exp(X * exponent)


TRENDS = {"polynomial": polynomial, "exponential": exponential}



from typing import Callable

import numpy as np
import pandas as pd
from scipy.optimize import minimize
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.metrics import mean_squared_error
from sklearn.utils.validation import check_is_fitted

from gtime.base import FeatureMixin, add_class_name
from gtime.utils.trends import TRENDS

__all__ = "Detrender"


class Detrender(BaseEstimator, TransformerMixin, FeatureMixin):
    """Apply a de-trend transformation to a time series.

    The purpose of the class is to fit a model, define through the `trend` parameter, in
    order to find a trend in the time series. Then, the trend can be removed by removing
    the predictions of the fitted model.

    Parameters
    ----------
    trend : ``'polynomial'`` | ``'exponential'``, required
        The kind of trend removal to apply.

    trend_x0 : np.array, required
        Initialisation parameters passed to the trend function. This is used to select
        a starting point in order to minimize the `loss` function.

    loss : Callable, optional, default: ``mean_squared_error``
        The loss function to minimize.

    method : string, optional, default: ``"BFGS"``
        Loss function optimisation method.

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from gtime.feature_extraction import Detrender
    >>> detrender = Detrender(trend='polynomial', trend_x0=np.zeros(2))
    >>> time_index = pd.date_range("2020-01-01", "2020-01-10")
    >>> X = pd.DataFrame(range(0, 10), index=time_index)
    >>> detrender.fit_transform(X)
                0__Detrender
    2020-01-01  9.180937e-07
    2020-01-02  8.020709e-07
    2020-01-03  6.860481e-07
    2020-01-04  5.700253e-07
    2020-01-05  4.540024e-07
    2020-01-06  3.379796e-07
    2020-01-07  2.219568e-07
    2020-01-08  1.059340e-07
    2020-01-09 -1.008878e-08
    2020-01-10 -1.261116e-07

    """

    def __init__(
        self,
        trend: str,
        trend_x0: np.array,
        loss: Callable = mean_squared_error,
        method: str = "BFGS",
    ):
        self.trend = trend
        self.trend_x0 = trend_x0
        self.loss = loss
        self.method = method

    def fit(self, X: pd.DataFrame, y=None) -> "Detrender":
        """Fit the estimator.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """

        # TODO: create validation function
        if self.trend not in TRENDS:
            raise ValueError(
                "The trend '%s' is not supported. Supported "
                "trends are %s." % (self.trend, list(sorted(TRENDS)))
            )

        self.best_trend_params_ = minimize(
            lambda opt: self.loss(
                X.values, [TRENDS[self.trend](t, opt) for t in range(0, X.shape[0])]
            ),
            self.trend_x0,
            method=self.method,
            options={"disp": False},
        )["x"]

        self.t0_ = X.index[0]
        freq = X.index.freq
        if freq is not None:
            self.period_ = freq
        else:
            self.period_ = X.index[1] - X.index[0]

        return self

    @add_class_name
    def transform(self, time_series: pd.DataFrame) -> pd.DataFrame:
        """Transform the ``time_series`` by removing the trend.

        Parameters
        ----------
        time_series: pd.DataFrame, shape (n_samples, 1), required
            The time series to transform.

        Returns
        -------
        time_series_t : pd.DataFrame, shape (n_samples, n_features)
            The transformed time series, without the trend.

        """
        check_is_fitted(self)

        time_steps = (time_series.index - self.t0_) / self.period_

        predictions = pd.Series(
            index=time_series.index,
            data=np.array(
                [TRENDS[self.trend](t, self.best_trend_params_) for t in time_steps]
            ).flatten(),
        )

        return time_series.sub(predictions, axis=0)



from datetime import datetime
from typing import List, Union, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.testing import assert_series_equal

from gtime.preprocessing.time_series_conversion import (
    _SequenceToTimeIndexSeries,
    _PandasSeriesToTimeIndexSeries,
    _TimeIndexSeriesToPeriodIndexSeries,
    count_not_none,
)
from gtime.utils.testing_constants import DEFAULT_START, DEFAULT_FREQ

PandasDate = Union[datetime, pd.Timestamp, str]


def compare_output_of_input_sequence_to_expected_one(
    input_sequence, start, end, freq,
):
    computed_pandas_series = transform_sequence_into_time_index_series(
        input_sequence, start, end, freq
    )
    expected_pandas_series = pandas_series_with_period_index(
        input_sequence, start, end, freq
    )
    assert_series_equal(computed_pandas_series, expected_pandas_series)


def compare_output_of_input_series_to_expected_one(
    input_sequence, start, end, freq,
):
    computed_pandas_series = transform_series_into_time_index_series(
        input_sequence, start, end, freq
    )
    expected_pandas_series = pandas_series_with_period_index(
        input_sequence.values, start, end, freq
    )
    assert_series_equal(computed_pandas_series, expected_pandas_series)


def transform_sequence_into_time_index_series(
    array_like_object: Union[np.array, list, pd.Series],
    start: Optional[str] = None,
    end: Optional[str] = None,
    freq: Optional[str] = None,
) -> pd.Series:
    time_series_conversion = _SequenceToTimeIndexSeries(start, end, freq)
    return time_series_conversion.transform(array_like_object)


def transform_series_into_time_index_series(
    array_like_object: Union[np.array, list, pd.Series],
    start: Optional[str] = None,
    end: Optional[str] = None,
    freq: Optional[str] = None,
) -> pd.Series:
    time_series_conversion = _PandasSeriesToTimeIndexSeries(start, end, freq)
    return time_series_conversion.transform(array_like_object)


def transform_time_index_series_into_period_index_series(
    series: pd.Series, freq: pd.Timedelta = None,
) -> pd.Series:
    to_period_conversion = _TimeIndexSeriesToPeriodIndexSeries(freq=freq)
    return to_period_conversion.transform(series)


def pandas_series_with_period_index(
    values: Union[np.array, List[float]],
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    freq: Optional[pd.Timedelta] = None,
) -> pd.Series:
    start, end, freq = _initialize_start_end_freq(start, end, freq)
    index = pd.period_range(start=start, end=end, periods=len(values), freq=freq,)
    return pd.Series(index=index, data=values, dtype=np.float64)


def _initialize_start_end_freq(
    start: PandasDate, end: PandasDate, freq: pd.Timedelta
) -> Tuple[pd.Timestamp, pd.Timestamp, pd.Timedelta]:
    not_none_params = count_not_none(start, end, freq)
    if not_none_params == 0:
        start, end, freq = _default_params_initialization()
    elif not_none_params == 1:
        start, end, freq = _one_not_none_param_initialization(start, end, freq)
    elif not_none_params == 2:
        start, end, freq = _two_not_none_params_initialization(start, end, freq)
    else:
        raise ValueError(
            "Of the three parameters: start, end, and "
            "freq, exactly two must be specified"
        )
    return start, end, freq


def _default_params_initialization() -> Tuple[pd.Timestamp, pd.Timestamp, pd.Timedelta]:
    start = DEFAULT_START
    end = None
    freq = DEFAULT_FREQ
    return start, end, freq


def _one_not_none_param_initialization(
    start, end, freq
) -> Tuple[pd.Timestamp, pd.Timestamp, pd.Timedelta]:
    if start is not None:
        start = start
        end = None
        freq = DEFAULT_FREQ
    elif end is not None:
        start = None
        end = end
        freq = DEFAULT_FREQ
    else:
        start = DEFAULT_START
        end = None
        freq = freq
    return start, end, freq


def _two_not_none_params_initialization(
    start, end, freq
) -> Tuple[pd.Timestamp, pd.Timestamp, pd.Timedelta]:
    start = start
    end = end
    freq = freq
    return start, end, freq


def datetime_index_series_to_period_index_series(
    datetime_index_series: pd.Series, freq: Optional[pd.Timedelta] = None
) -> pd.Series:
    if datetime_index_series.index.freq is not None:
        try:
            return pd.Series(
                index=pd.PeriodIndex(datetime_index_series.index),
                data=datetime_index_series.values,
            )
        except Exception as e:
            print(freq, datetime_index_series.index.freq)
            raise e
    else:
        freq = "1D" if freq is None else freq
        return pd.Series(
            index=pd.PeriodIndex(datetime_index_series.index, freq=freq),
            data=datetime_index_series.values,
        )


def timedelta_index_to_datetime(
    index: pd.TimedeltaIndex, start: datetime = datetime(year=1970, month=1, day=1),
) -> pd.DatetimeIndex:
    return start + index


def timedelta_index_series_to_period_index_series(
    timedelta_index_series: pd.Series, freq: Optional[pd.Timedelta] = None
) -> pd.Series:
    datetime_index = timedelta_index_to_datetime(timedelta_index_series.index)
    if datetime_index.freq is None:
        freq = "1D" if freq is None else freq
        period_index = pd.PeriodIndex(datetime_index, freq=freq)
    else:
        period_index = pd.PeriodIndex(datetime_index)
    return pd.Series(index=period_index, data=timedelta_index_series.values)



import pandas as pd
from sklearn.compose import ColumnTransformer


class FeatureCreation(ColumnTransformer):
    """Applies transformers to columns of a pandas DataFrame.

    This estimator is a wrapper of sklearn.compose.ColumnTransformer, the only
    difference is the output type of fit_transform and transform methods which is a
    DataFrame instead of an array.

    """

    def fit_transform(self, X: pd.DataFrame, y: pd.DataFrame = None):
        """Fit all transformers, transform the data and concatenate results.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            Input data, of which specified subsets are used to fit the
            transformers.

        y : pd.DataFrame, shape (n_samples, ...), optional, default: ``None``
            Targets for supervised learning.

        Examples
        --------
        >>> import pandas.util.testing as testing
        >>> from gtime.compose import FeatureCreation
        >>> from gtime.feature_extraction import Shift, MovingAverage
        >>> data = testing.makeTimeDataFrame(freq="s")
        >>> fc = FeatureCreation([
        ...     ('s1', Shift(1), ['A']),
        ...     ('ma3', MovingAverage(window_size=3), ['B']),
        ... ])
        >>> fc.fit_transform(data).head()
                             s1__A__Shift  ma3__B__MovingAverage
        2000-01-01 00:00:00           NaN                    NaN
        2000-01-01 00:00:01      0.211403                    NaN
        2000-01-01 00:00:02     -0.313854               0.085045
        2000-01-01 00:00:03      0.502018              -0.239269
        2000-01-01 00:00:04     -0.225324              -0.144625

        Returns
        -------
        X_t_df : pd.DataFrame, shape (n_samples, sum_n_components)
            hstack of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers.

        """
        X_t = super().fit_transform(X, y)
        X_t_df = pd.DataFrame(data=X_t, columns=self.get_feature_names(), index=X.index)
        return X_t_df

    def transform(self, X: pd.DataFrame):
        """Transform X separately by each transformer, concatenate results.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            The data to be transformed by subset.

        Returns
        -------
        X_t_df : DataFrame, shape (n_samples, sum_n_components)
            hstack of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers. If
            any result is a sparse matrix, everything will be converted to
            sparse matrices.

        """
        X_t = super().transform(X)
        X_t_df = pd.DataFrame(data=X_t, columns=self.get_feature_names(), index=X.index)
        return X_t_df



from typing import Callable

import numpy as np
import pandas as pd
from scipy.optimize import minimize
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.metrics import mean_squared_error
from sklearn.utils.validation import check_is_fitted

from gtime.utils.trends import TRENDS


class TrendForecaster(BaseEstimator, RegressorMixin):
    """Trend forecasting model.

    This estimator optimizes a trend function on train data and will forecast using this trend function with optimized
    parameters.

    Parameters
    ----------
    trend : ``"polynomial"`` | ``"exponential"``, required
        The kind of trend removal to apply.

    trend_x0 : np.array, required
        Initialisation parameters passed to the trend function

    loss : Callable, optional, default: ``mean_squared_error``
        Loss function to minimize.

    method : str, optional, default: ``"BFGS"``
        Loss function optimisation method

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from gtime.model_selection import horizon_shift, FeatureSplitter
    >>> from gtime.forecasting import TrendForecaster
    >>>
    >>> X = pd.DataFrame(np.random.random((10, 1)), index=pd.date_range("2020-01-01", "2020-01-10"))
    >>> y = horizon_shift(X, horizon=2)
    >>> X_train, y_train, X_test, y_test = FeatureSplitter().transform(X, y)
    >>>
    >>> tf = TrendForecaster(trend='polynomial', trend_x0=np.zeros(2))
    >>> tf.fit(X_train).predict(X_test)
    array([[0.39703029],
           [0.41734957]])

    """

    def __init__(
        self,
        trend: str,
        trend_x0: np.array,
        loss: Callable = mean_squared_error,
        method: str = "BFGS",
    ):
        self.trend = trend
        self.trend_x0 = trend_x0
        self.loss = loss
        self.method = method

    def fit(self, X: pd.DataFrame, y=None) -> "TrendForecaster":
        """Fit the estimator.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """

        if self.trend not in TRENDS:
            raise ValueError(
                "The trend '%s' is not supported. Supported "
                "trends are %s." % (self.trend, list(sorted(TRENDS)))
            )

        self.best_trend_params_ = minimize(
            lambda opt: self.loss(
                X.values, [TRENDS[self.trend](t, opt) for t in range(0, X.shape[0])]
            ),
            self.trend_x0,
            method=self.method,
            options={"disp": False},
        )["x"]

        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Using the fitted polynomial, predict the values starting from ``X``.

        Parameters
        ----------
        X: pd.DataFrame, shape (n_samples, 1), required
            The time series on which to predict.

        Returns
        -------
        predictions : pd.DataFrame, shape (n_samples, 1)
            The output predictions.

        Raises
        ------
        NotFittedError
            Raised if the model is not fitted yet.

        """
        check_is_fitted(self)

        predictions = TRENDS[self.trend](X.values, self.best_trend_params_)
        return predictions



import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given, settings, HealthCheck
from sklearn.compose import make_column_selector

from gtime.compose import FeatureCreation
from gtime.feature_extraction import Shift, MovingAverage
from gtime.model_selection.splitters import FeatureSplitter
from gtime.utils.hypothesis.feature_matrices import X_y_matrices

# TODO: refactor, make hypothesis generator instead of a full pipeline
df_transformer = FeatureCreation(
    [
        ("shift_0", Shift(0), make_column_selector(dtype_include=np.number)),
        ("shift_1", Shift(1), make_column_selector(dtype_include=np.number)),
        (
            "moving_average_3",
            MovingAverage(window_size=3),
            make_column_selector(dtype_include=np.number),
        ),
    ]
)

horizon = 4


class TestFeatureSplitter:
    def test_constructor(self):
        FeatureSplitter()

    @given(st.text().filter(lambda x: x != "any"))
    def test_constructor_wrong_parameter(self, drop_na_mode: str):
        with pytest.raises(ValueError):
            FeatureSplitter(drop_na_mode)

    @settings(suppress_health_check=(HealthCheck.too_slow,))
    @given(
        X_y_matrices(
            horizon=horizon, df_transformer=df_transformer, allow_nan_infinity=False,
        )
    )
    def test_transform(self, X_y):
        X, y = X_y
        feature_splitter = FeatureSplitter()
        X_train, y_train, X_test, y_test = feature_splitter.transform(X, y)

        assert X_train.shape[0] == max(0, X.shape[0] - 2 - horizon)
        assert y_train.shape[0] == X_train.shape[0]
        assert X_test.shape[0] == min(max(0, X.shape[0] - 2), horizon)
        assert y_test.shape[0] == X_test.shape[0]



from datetime import datetime
from typing import Union, Tuple

import hypothesis.strategies as st
import pandas as pd


def initialize_start_date_end_date(
    start: datetime, end: datetime
) -> Tuple[datetime, datetime]:
    start = start if start is not None else pd.Timestamp("1980-01-01")
    end = end if end is not None else pd.Timestamp("2020-01-01")
    return start, end


def initialize_start_timedelta_end_timedelta(start: pd.Timedelta, end: pd.Timedelta):
    start = start if start is not None else pd.Timedelta(0)
    end = end if end is not None else pd.Timedelta("40Y")
    return start, end


def order_pair(element1, element2):
    return st.builds(
        lambda start, end: (start, end), start=element1, end=element2
    ).filter(lambda x: x[0] < x[1])


def expected_start_date_from(
    end: Union[datetime, pd.Period], periods: int, freq: pd.Timedelta
) -> Union[datetime, pd.Period]:
    return end - periods * freq


def expected_end_date_from(
    start: Union[datetime, pd.Period], periods: int, freq: pd.Timedelta
) -> Union[datetime, pd.Period]:
    return start + periods * freq


def expected_index_length_from(
    start: Union[datetime, pd.Period],
    end: Union[datetime, pd.Period],
    freq: pd.Timedelta,
) -> int:
    expected_index_length = (end - start) // freq
    return expected_index_length


def freq_to_timedelta(
    freq: str, approximate_if_non_uniform: bool = True
) -> pd.Timedelta:
    try:
        return pd.to_timedelta(f"1{freq}")
    except ValueError as e:
        if approximate_if_non_uniform:
            correspondences = {
                "B": pd.Timedelta(1, unit="D"),
                "Q": pd.Timedelta(90, unit="D"),
                "A": pd.Timedelta(365, unit="D"),
            }
            return correspondences[freq]
        else:
            raise e



import random

import numpy as np
import pandas as pd
if pd.__version__ >= '1.0.0':
    import pandas._testing as testing
else:
    import pandas.util.testing as testing

import pytest
from hypothesis import given
from sklearn.compose import make_column_selector
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression

from gtime.compose import FeatureCreation
from gtime.feature_extraction import MovingAverage, Shift
from gtime.feature_generation import Constant
from gtime.forecasting import GAR, GARFF
from gtime.model_selection import FeatureSplitter
from gtime.utils.hypothesis.feature_matrices import X_y_matrices

df_transformer = FeatureCreation(
    [
        ("shift_0", Shift(0), make_column_selector(dtype_include=np.number)),
        ("shift_1", Shift(1), make_column_selector(dtype_include=np.number)),
        (
            "moving_average_3",
            MovingAverage(window_size=3),
            make_column_selector(dtype_include=np.number),
        ),
    ]
)


@pytest.fixture
def time_series():
    testing.N, testing.K = 200, 1
    return testing.makeTimeDataFrame(freq="MS")


def arbitrary_features(feature_length):
    possible_features = [MovingAverage, Constant, Shift]
    random_features = []
    random_params = random.sample(range(1, 100), feature_length)

    feature_names = []
    for random_param in random_params:
        random_feature = random.sample(possible_features, 1)[0]
        output_name = f"{random_feature}_{random_params}"
        if output_name in feature_names:
            continue
        feature_names.append(output_name)
        feature = random_feature(random_param)
        random_features.append(feature)

    return random_features, feature_names


class TestFitPredict:
    def test_predict_with_no_fit(self, time_series):
        base_model = LinearRegression()

        gar_no_feedforward = GAR(estimator=base_model)

        with pytest.raises(NotFittedError):
            gar_no_feedforward.predict(time_series)

    @given(
        X_y_matrices(
            horizon=4,
            df_transformer=df_transformer,
            allow_nan_infinity=False,
            min_length=10,
        )
    )
    def test_correct_fit_date(self, X_y):
        base_model = LinearRegression()
        feature_splitter = FeatureSplitter()
        x, y = X_y[0], X_y[1]
        x_train, y_train, x_test, y_test = feature_splitter.transform(x, y)

        gar_no_feedforward = GAR(estimator=base_model)

        gar_no_feedforward.fit(x_train, y_train)

        predictions = gar_no_feedforward.predict(x_test)

        assert len(predictions) == len(x_test)
        np.testing.assert_array_equal(predictions.index, x_test.index)

        gar_with_feedforward = GARFF(estimator=base_model)

        gar_with_feedforward.fit(x_train, y_train)

        predictions = gar_with_feedforward.predict(x_test)

        assert len(predictions) == len(x_test)
        np.testing.assert_array_equal(predictions.index, x_test.index)



from math import pi
from typing import Union, Optional

import numpy as np
import pandas as pd
from pandas import DatetimeIndex
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from gtime.base import FeatureMixin, add_class_name

__all__ = ["PeriodicSeasonal", "Constant"]


# TODO: add something like 'make_periodic_feature' or 'make_sinusoid'
class PeriodicSeasonal(BaseEstimator, TransformerMixin, FeatureMixin):
    """Create a sinusoid from a given date and with a given period and amplitude.

    Parameters
    ----------
    period : Union[pd.Timedelta, str], optional, default: ``'365 days'``
        The period of the generated time series.

    amplitude : float, optional, default: ``0.5``
        The amplitude of the time series.

    start_date : Union[pd.Timestamp, str], optional, default: ``None``
        The date from which to start generating the feature. This is used only if X is
        not passed in the ``transform`` method, otherwise the start date is inferred
        from it.

    length : int, optional, default: ``50``
        The length of the sinusoid. This is used only if X is not passed in the
        ``transform`` method, otherwise the length is inferred from it.

    index_period : Union[DatetimeIndex, int], optional, default: ``None``
        The period of the index of the output ``DataFrame``. This is used only if X is
        not passed in the ``transform`` method, otherwise the index period is taken
        from it.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_generation import PeriodicSeasonal
    >>> X = pd.DataFrame(range(0, 10), index=pd.date_range(start='2019-04-18',  end='2019-04-27', freq='d'))
    >>> periodic = PeriodicSeasonal()
    >>> periodic.fit_transform(X)
                0__PeriodicSeasonal
    2019-04-18             0.000000
    2019-04-19             0.008607
    2019-04-20             0.017211
    2019-04-21             0.025810
    2019-04-22             0.034401
    2019-04-23             0.042982
    2019-04-24             0.051551
    2019-04-25             0.060104
    2019-04-26             0.068639
    2019-04-27             0.077154

    """

    def __init__(
        self,
        period: Union[pd.Timedelta, str] = "365 days",
        amplitude: float = 0.5,
        start_date: Optional[Union[pd.Timestamp, str]] = None,
        length: Optional[int] = 50,
        index_period: Optional[Union[DatetimeIndex, int]] = None,
    ):
        self.start_date = start_date
        self.period = period
        self.amplitude = amplitude
        self.length = length
        self.index_period = index_period

    def fit(self, X: pd.DataFrame, y=None) -> "PeriodicSeasonal":
        """Fit the estimator.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features), required
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : PeriodicSeasonal
            Returns self.

        """
        self.columns_ = X.columns.values
        return self

    @add_class_name
    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a sinusoid, with the given ``period``, ``amplitude`` and ``length``,
        starting from the selected ``start_date``. If ``time_series`` is not ``None``,
        the ``start_date`` is replaced by the start date of the time series and the
        output sinusoid will have the same index as ``time_series``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            The input DataFrame, If passed, the output DataFrame is going to have the
            same index as ``time_series``. If is not passed, then the ``start_date`` and
            ``index_period`` must have been passed in the constructor when the object
            was instantiated.

        Returns
        -------
        periodic_feature : pd.DataFrame, shape (n_samples, 1)
            The DataFrame containing the generated period feature.

        Raises
        ------
        ValueError
            Raised if ``time_series`` is not passed and the ``start_date`` or the
            ``index_period`` are not present.

        """
        self._validate_input(time_series)
        self._transform_inputs(time_series)
        datetime_index = self._get_time_index(time_series)

        periodic_feature_values = self._compute_periodic_feature(datetime_index)
        if time_series is not None:
            periodic_feature = pd.DataFrame(
                index=time_series.index, data=periodic_feature_values
            )

        else:
            periodic_feature = pd.DataFrame(data=periodic_feature_values[: self.length])

        return periodic_feature

    def _validate_input(self, X: pd.DataFrame) -> None:
        if X is None:
            if self.start_date is None:
                raise ValueError(
                    f"If X is not provided, the 'start_date' variable "
                    f"must be provided, but instead was "
                    f"{self.start_date}."
                )
            if self.index_period is None:
                raise ValueError(
                    f"If X is not provided, the 'index_period' variable "
                    f"must be provided, but instead was "
                    f"{self.index_period}."
                )

    def _transform_inputs(self, X: pd.DataFrame) -> None:
        if isinstance(self.period, str):
            self.period = pd.to_timedelta(self.period)

        if X is not None:
            self.start_date = X.index.values[0]
        else:
            if isinstance(self.start_date, str):
                self.start_date = pd.to_datetime(self.start_date)

    def _get_time_index(self, X: pd.DataFrame) -> pd.DatetimeIndex:
        if X is not None:
            datetime_index = X.index
        else:
            if isinstance(self.index_period, int):
                datetime_index = pd.date_range(
                    start=self.start_date, periods=self.index_period
                )
            else:
                datetime_index = self.index_period

        self._check_sampling_frequency(datetime_index)

        return datetime_index

    def _check_sampling_frequency(self, datetime_index: pd.DatetimeIndex) -> None:
        if isinstance(datetime_index, pd.PeriodIndex):
            datetime_index = datetime_index.to_timestamp()
        sampling_frequency = pd.Timedelta(datetime_index[1] - datetime_index[0])

        if self.period < 2 * sampling_frequency:
            raise ValueError(
                f"Sampling frequency must be at least two times"
                f"the period to obtain meaningful results. "
                f"Sampling frequency = {sampling_frequency},"
                f"period = {self.period}."
            )

    def _compute_periodic_feature(self, datetime_index: pd.DatetimeIndex):
        return (
            np.sin(2 * pi * (datetime_index - self.start_date) / self.period)
        ) * self.amplitude


class Constant(BaseEstimator, TransformerMixin, FeatureMixin):
    """Generate a ``pd.DataFrame`` with one column, of the same length as the input
    ``X`` and containing the value ``constant`` across the whole column.

    Parameters
    ----------
    constant : int, optional, default: ``2``
        The value to use to generate the constant column of the ``pd.DataFrame``.

    length : int, optional, default: ``50``
        The length of the DataFrame to generate. This is used only if X is not passed in
        the ``transform`` method, otherwise the length is inferred from it.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_generation import Constant
    >>> X = pd.DataFrame(range(0, 5), index=pd.date_range(start='2019-04-18',  end='2019-04-22', freq='d'))
    >>> constant = Constant(constant=3)
    >>> constant.fit_transform(X)
                0__Constant
    2019-04-18          3.0
    2019-04-19          3.0
    2019-04-20          3.0
    2019-04-21          3.0
    2019-04-22          3.0

    """

    def __init__(self, constant: int = 0, length: int = None):
        super().__init__()
        self.length = length
        self.constant = constant

    def get_feature_names(self):
        """Return feature names for output features.

        Returns
        -------
        output_feature_names : ndarray, shape (n_output_features,)
            Array of feature names.

        """

        return [self.__class__.__name__]

    def fit(self, X: pd.DataFrame, y=None) -> "Constant":
        """Fit the estimator.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : Constant
            Returns self.

        """
        self.columns_ = X.columns.values
        self.rows_ = X.shape[0]
        self.length_ = (
            np.min([self.length, self.rows_]) if self.length is not None else self.rows_
        )
        return self

    @add_class_name
    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a ``pd.DataFrame`` with one column with the same length as
        ``time_series`` and with the same index, containing a value equal to
        ``constant``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            The input DataFrame. If passed, the output DataFrame is going to have the
            same index as ``time_series``.

        Returns
        -------
        constant_series_renamed : pd.DataFrame, shape (length, 1)
            A constant series, with the same length of ``X`` and with the same index.

        """
        check_is_fitted(self)

        constants = np.full(self.rows_, np.nan)
        constants[: self.length_] = self.constant
        if time_series is not None:
            constant_series = pd.Series(
                data=constants, index=time_series.index
            ).to_frame()
        else:
            constant_series = pd.Series(data=constants).to_frame()

        return constant_series



from random import randint

import numpy as np
import pytest
from hypothesis import given, strategies as st
from pandas.util import testing as testing

from gtime.causality import ShiftedLinearCoefficient
from gtime.causality.tests.common import make_df_from_expected_shifts


def test_linear_coefficient():
    expected_shifts = [randint(2, 9) * 2 for _ in range(3)]

    df = make_df_from_expected_shifts(expected_shifts)
    slc = ShiftedLinearCoefficient(target_col="A", max_shift=20)
    slc.fit(df)

    shifts = slc.best_shifts_["A"][1:].values
    np.testing.assert_array_equal(shifts, expected_shifts)


# TODO: tests refactor TBD
@given(st.integers(1, 20))
@pytest.mark.skip(reason="TODO: Write proper test, increase hypothesis max duration")
def test_linear_coefficient_hyp(shift):
    testing.N, testing.K = 500, 1
    df = testing.makeTimeDataFrame(freq="D")
    df["shifted"] = df["A"].shift(shift)

    slc = ShiftedLinearCoefficient(target_col="A", max_shift=20)
    slc.fit(df).transform(df)


def test_linear_bootstrap_p_values():
    # This test and the next one just test if the p_values on the diagonal are equal
    # to 0. Is hard to implement other unittest, since the bootstrapping always
    # gives different result. However, other properties could be tested
    expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
    df = make_df_from_expected_shifts(expected_shifts)
    shifted_test = ShiftedLinearCoefficient(
        target_col="A", max_shift=5, bootstrap_iterations=500,
    )
    shifted_test.fit(df)

    linear_p_values = shifted_test.bootstrap_p_values_
    for col_index in range(len(linear_p_values.columns)):
        assert linear_p_values.iloc[col_index, col_index] == 0


def test_linear_permutation_p_values():
    expected_shifts = [randint(2, 9) * 2 for _ in range(3)]
    df = make_df_from_expected_shifts(expected_shifts)
    shifted_test = ShiftedLinearCoefficient(
        target_col="A", max_shift=5, permutation_iterations=50,
    )
    shifted_test.fit(df)

    linear_p_values = shifted_test.permutation_p_values_
    for col_index in range(len(linear_p_values.columns)):
        assert linear_p_values.iloc[col_index, col_index] == 0



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.linear_model import LinearRegression

from gtime.causality.base import CausalityMixin


class ShiftedLinearCoefficient(BaseEstimator, TransformerMixin, CausalityMixin):
    """Test the shifted linear fit coefficients between two or more time series.

    Parameters
    ----------
    min_shift : int, optional, default: ``1``
        The minimum number of shifts to check for.

    max_shift : int, optional, default: ``10``
        The maximum number of shifts to check for.

    target_col : str, optional, default: ``None``
        The column to use as the a reference (i.e., the column which is not
        shifted).

    dropna : bool, optional, default: ``False``
        Determines if the Nan values created by shifting are retained or dropped.

    bootstrap_iterations : int, optional, default: ``None``
        If not None, compute the p_values of the test, by performing bootstrapping of
        the original data (sampling with replacement).

    permutation_iterations : int, optional, default: ``None``
        If not None, compute the p_values of the test, by performing permutations of
        the original data.

    Examples
    --------

    >>> from gtime.causality.linear_coefficient import ShiftedLinearCoefficient
    >>> import pandas.util.testing as testing
    >>> data = testing.makeTimeDataFrame(freq="s")
    >>> slc = ShiftedLinearCoefficient(target_col="A")
    >>> slc.fit(data)
    >>> slc.best_shifts_
    y  A  B  C  D
    x
    A  3  6  8  5
    B  9  9  4  1
    C  8  2  4  9
    D  3  9  4  3
    >>> slc.max_corrs_
    y         A         B         C         D
    x
    A  0.460236  0.420005  0.339370  0.267143
    B  0.177856  0.300350  0.367150  0.550490
    C  0.484860  0.263036  0.456046  0.251342
    D  0.580068  0.344688  0.253626  0.256220
    """

    def __init__(
        self,
        min_shift: int = 1,
        max_shift: int = 10,
        target_col: str = None,
        dropna: bool = False,
        bootstrap_iterations: int = None,
        permutation_iterations: int = None,
    ):
        super().__init__(
            bootstrap_iterations=bootstrap_iterations,
            permutation_iterations=permutation_iterations,
        )
        self.min_shift = min_shift
        self.max_shift = max_shift
        self.target_col = target_col
        self.dropna = dropna

    def fit(self, data: pd.DataFrame) -> "ShiftedLinearCoefficient":
        """Create the DataFrame of shifts of each time series which maximize the shifted
         linear fit coefficients.

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time-series on which to compute the shifted
            linear fit coefficients.

        Returns
        -------
        self : ``ShiftedLinearCoefficient``

        """
        best_shifts = self._compute_best_shifts(data, self._get_max_coeff_shift)
        pivot_tables = self._create_pivot_tables(best_shifts)

        self.best_shifts_ = pivot_tables["best_shifts"]
        self.max_corrs_ = pivot_tables["max_corrs"]

        if self.bootstrap_iterations:
            self.bootstrap_p_values_ = pivot_tables["bootstrap_p_values"]

        if self.permutation_iterations:
            self.permutation_p_values_ = pivot_tables["permutation_p_values"]

        return self

    def _get_max_coeff_shift(self, data: pd.DataFrame, x, y):
        shifts = pd.DataFrame()

        shifts[x] = data[x]
        shifts[y] = data[y]

        for shift in range(self.min_shift, self.max_shift + 1):
            shifts[shift] = data[x].shift(shift)

        shifts = shifts.dropna()
        lf = LinearRegression().fit(
            shifts[range(self.min_shift, self.max_shift + 1)].values, shifts[y].values
        )

        q = lf.coef_.max(), np.argmax(lf.coef_) + (self.min_shift - 0)
        return q



import numpy as np
import pandas as pd
import pytest
from scipy.stats import gmean
from hypothesis import given
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import floats

from gtime.metrics import (
    smape,
    max_error,
    mse,
    log_mse,
    r_square,
    mae,
    mape,
    rmse,
    rmsle,
    gmae,
)


class TestSmape:
    def _correct_smape(self, y_true, y_pred):
        non_normalized_smape = sum(
            np.abs(y_pred - y_true) / (np.abs(y_pred) + np.abs(y_true))
        )
        non_normalized_smape_filled = np.nan_to_num(non_normalized_smape, nan=0)
        error = (2 / len(y_pred)) * non_normalized_smape_filled
        return error

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            smape(y_true, y_pred)

    def test_nan_values(self):
        y_true = np.array([np.nan, 1, 2, 3])
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            smape(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = np.array([0, np.inf, 2, 3])

        with pytest.raises(ValueError):
            smape(y_true, y_pred)

    def test_correct_smape_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([3, 2, 4, 3, 5, 7])

        error = smape(y_true, y_pred)
        expected_error = (1 + 1 / 3 + 1 / 3 + 0 + 1 / 9 + 1 / 6) * (2 / 6)

        assert expected_error == error

    def test_correct_smape_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([3, 2, 4, 3, 5, 7])

        error = smape(y_true, y_pred)
        expected_error = (1 + 1 / 3 + 1 / 3 + 0 + 1 / 9 + 1 / 6) * (2 / 6)

        assert expected_error == error

    def test_correct_smape_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [3, 2, 4, 3, 5, 7]

        error = smape(y_true, y_pred)
        expected_error = (1 + 1 / 3 + 1 / 3 + 0 + 1 / 9 + 1 / 6) * (2 / 6)

        assert expected_error == error

    def test_smape_is_symmetric(self):
        y_true = np.random.random(10)
        y_pred = np.random.random(10)

        first_error = smape(y_true, y_pred)
        second_error = smape(y_pred, y_true)

        assert first_error == second_error

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_between_zero_and_one(self, y_true, y_pred):
        error = smape(y_true, y_pred)

        assert error >= 0
        assert error <= 2

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_smape_random_arrays(self, y_true, y_pred):
        error = smape(y_true, y_pred)
        expected_error = self._correct_smape(y_true, y_pred)

        assert expected_error == error


class TestMaxError:
    def _correct_max_error(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        error = np.amax(np.absolute((y_true - y_pred)))
        return error

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            max_error(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            max_error(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            max_error(y_true, y_pred)

    def test_max_error_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        error = max_error(y_true, y_pred)
        expected_max_error = 7

        assert expected_max_error == error

    def test_max_error_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        error = max_error(y_true, y_pred)
        expected_max_error = 7

        assert expected_max_error == error

    def test_max_error_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        error = max_error(y_true, y_pred)
        expected_max_error = 7

        assert expected_max_error == error

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_max_error_random_arrays_finite_values(self, y_true, y_pred):
        error = max_error(y_true, y_pred)
        expected_error = self._correct_max_error(y_true, y_pred)
        print(y_true)
        print(y_pred)
        assert expected_error == error


class TestMSE:
    def _correct_mse(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        sum_squared_error = sum(((y_true - y_pred)) ** 2)
        mse = sum_squared_error / len(y_true)
        return mse

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            mse(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            mse(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            mse(y_true, y_pred)

    def test_mse_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        mse_value = np.round(mse(y_true, y_pred), decimals=2)
        expected_mse = 14

        assert expected_mse == mse_value

    def test_mse_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        mse_value = np.round(mse(y_true, y_pred), decimals=2)
        expected_mse = 14

        assert expected_mse == mse_value

    def test_mse_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        mse_value = np.round(mse(y_true, y_pred), decimals=2)
        expected_mse = 14

        assert expected_mse == mse_value

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_mse_random_arrays_finite_values(self, y_true, y_pred):
        mse_value = mse(y_true, y_pred)
        expected_mse = self._correct_mse(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_mse == mse_value


class TestRMSE:
    def _correct_rmse(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        return np.sqrt(mse(y_true, y_pred))

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            rmse(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            rmse(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            rmse(y_true, y_pred)

    def test_rmse_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        rmse_value = np.round(rmse(y_true, y_pred), decimals=2)
        expected_rmse = 3.74

        assert expected_rmse == rmse_value

    def test_rmse_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        rmse_value = np.round(rmse(y_true, y_pred), decimals=2)
        expected_rmse = 3.74

        assert expected_rmse == rmse_value

    def test_rmse_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        rmse_value = np.round(rmse(y_true, y_pred), decimals=2)
        expected_rmse = 3.74

        assert expected_rmse == rmse_value

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_rmse_random_arrays_finite_values(self, y_true, y_pred):
        rmse_value = rmse(y_true, y_pred)
        expected_rmse = self._correct_rmse(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_rmse == rmse_value


class TestLogMSE:
    def _correct_log_mse(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        if (np.any(y_true < 0)) or (np.any(y_pred < 0)):
            raise ValueError("MSLE can not be used when inputs contain Negative values")
        log_y_true = np.log(y_true + 1)
        log_y_pred = np.log(y_pred + 1)
        log_mse = mse(log_y_true, log_y_pred)
        return log_mse

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            log_mse(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            log_mse(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            log_mse(y_true, y_pred)

    def test_log_mse_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [1, 4, 5, 10, 4, 1]

        log_mse_value = np.round(log_mse(y_true, y_pred), decimals=2)
        expected_log_mse = 0.67

        assert expected_log_mse == log_mse_value

    def test_log_mse_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([1, 4, 5, 10, 4, 1])

        log_mse_value = np.round(log_mse(y_true, y_pred), decimals=2)
        expected_log_mse = 0.67

        assert expected_log_mse == log_mse_value

    def test_log_mse_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([1, 4, 5, 10, 4, 1])

        log_mse_value = np.round(log_mse(y_true, y_pred), decimals=2)
        expected_log_mse = 0.67

        assert expected_log_mse == log_mse_value

    def test_log_mse_negative_values(self):
        y_true = np.array([0, 1, 2, 3, 4, -5])
        y_pred = np.array([0, 0, 0, 0, 0, 0])

        with pytest.raises(ValueError):
            log_mse(y_true, y_pred)

    @given(
        arrays(
            float,
            shape=30,
            elements=floats(allow_nan=False, allow_infinity=False, min_value=0),
        ),
        arrays(
            float,
            shape=30,
            elements=floats(allow_nan=False, allow_infinity=False, min_value=0),
        ),
    )
    def test_log_mse_random_arrays_finite_values(self, y_true, y_pred):
        log_mse_value = log_mse(y_true, y_pred)
        expected_log_mse = self._correct_log_mse(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_log_mse == log_mse_value


class TestRMSLE:
    def _correct_rmsle(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)

        return np.sqrt(log_mse(y_true, y_pred))

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            rmsle(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            rmsle(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            rmsle(y_true, y_pred)

    def test_rmsle_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [1, 4, 5, 10, 4, 1]

        rmsle_value = np.round(rmsle(y_true, y_pred), decimals=2)
        expected_rmsle = 0.82

        assert expected_rmsle == rmsle_value

    def test_rmsle_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([1, 4, 5, 10, 4, 1])

        rmsle_value = np.round(rmsle(y_true, y_pred), decimals=2)
        expected_rmsle = 0.82

        assert expected_rmsle == rmsle_value

    def test_rmsle_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([1, 4, 5, 10, 4, 1])

        rmsle_value = np.round(rmsle(y_true, y_pred), decimals=2)
        expected_rmsle = 0.82

        assert expected_rmsle == rmsle_value

    def test_rmsle_negative_values(self):
        y_true = np.array([0, 1, 2, 3, 4, -5])
        y_pred = np.array([0, 0, 0, 0, 0, 0])

        with pytest.raises(ValueError):
            rmsle(y_true, y_pred)

    @given(
        arrays(
            float,
            shape=30,
            elements=floats(allow_nan=False, allow_infinity=False, min_value=0),
        ),
        arrays(
            float,
            shape=30,
            elements=floats(allow_nan=False, allow_infinity=False, min_value=0),
        ),
    )
    def test_rmsle_random_arrays_finite_values(self, y_true, y_pred):
        rmsle_value = rmsle(y_true, y_pred)
        expected_rmsle = self._correct_rmsle(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_rmsle == rmsle_value


class TestRSquare:
    def _correct_r_squared(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        ss_res = sum(((y_true - y_pred)) ** 2)
        ss_tot = sum((y_true - np.mean(y_true)) ** 2)
        if not np.any(ss_tot):
            if not np.any(ss_res):
                return 1.0
            else:
                return 0.0
        if np.isnan(ss_res / ss_tot):
            return np.NINF
        r_square = 1 - (ss_res / ss_tot)
        return r_square

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            r_square(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            r_square(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            r_square(y_true, y_pred)

    def test_r_square_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        r_square_value = np.round(r_square(y_true, y_pred), decimals=2)
        expected_r_square = -3.8

        assert expected_r_square == r_square_value

    def test_r_square_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        r_square_value = np.round(r_square(y_true, y_pred), decimals=2)
        expected_r_square = -3.8

        assert expected_r_square == r_square_value

    def test_r_square_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        r_square_value = np.round(r_square(y_true, y_pred), decimals=2)
        expected_r_square = -3.8

        assert expected_r_square == r_square_value

    def test_r_square_all_zero_values(self):
        # The test checks for all zero inputs covering the 'if' condition
        # to check if both ss_res and ss_tot are all Zeros
        y_true = np.array([0, 0, 0, 0, 0, 0])
        y_pred = np.array([0, 0, 0, 0, 0, 0])

        r_square_value = np.round(r_square(y_true, y_pred), decimals=2)
        expected_r_square = 1.0

        assert expected_r_square == r_square_value

    def test_r_square_ss_tot_zero(self):
        # The test checks for 'if' condition
        # to check if ss_res is certain value and ss_tot are all Zeros
        y_true = np.array([0, 0, 0, 0, 0, 0])
        y_pred = np.array([1, 1, 1, 1, 1, 1])

        r_square_value = np.round(r_square(y_true, y_pred), decimals=2)
        expected_r_square = 0.0

        assert expected_r_square == r_square_value

    def test_r_square_ss_res_ss_tot_infinity(self):
        # The test checks for 'if' condition
        # to check if both ss_res and ss_tot are Infinite
        y_true = [
            0.00000000e000,
            0.00000000e000,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
            9.81351055e200,
        ]
        y_pred = [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]

        r_square_value = np.round(r_square(y_true, y_pred), decimals=2)
        expected_r_square = np.NINF

        assert expected_r_square == r_square_value

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_r_square_random_arrays_finite_values(self, y_true, y_pred):
        r_square_value = r_square(y_true, y_pred)
        expected_r_square = self._correct_r_squared(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_r_square == r_square_value


class TestMAE:
    def _correct_mae(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        mae = np.mean(np.abs(y_true - y_pred))
        return mae

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            mae(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            mae(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            mae(y_true, y_pred)

    def test_mae_list(self):
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        mae_value = np.round(mae(y_true, y_pred), decimals=2)
        expected_mae = 3

        assert expected_mae == mae_value

    def test_mae_array(self):
        y_true = np.array([0, 1, 2, 3, 4, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        mae_value = np.round(mae(y_true, y_pred), decimals=2)
        expected_mae = 3

        assert expected_mae == mae_value

    def test_mae_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        mae_value = np.round(mae(y_true, y_pred), decimals=2)
        expected_mae = 3

        assert expected_mae == mae_value

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_mae_random_arrays_finite_values(self, y_true, y_pred):
        mae_value = mae(y_true, y_pred)
        expected_mae = self._correct_mae(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_mae == mae_value


class TestMAPE:
    def _correct_mape(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        mape_value = np.mean(np.abs((y_pred - y_true) / y_true))
        if np.isnan(mape_value):
            raise ValueError("MAPE can not be calculated due to Zero/Zero")
        return mape_value * 100

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            mape(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            mape(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            mape(y_true, y_pred)

    def test_mape_list(self):
        y_true = [1, 5, 7, 3, 4, 5]
        y_pred = [-1, 3, 5, 4, 4, 3]

        mape_value = np.round(mape(y_true, y_pred), decimals=2)
        expected_mape = 56.98

        assert expected_mape == mape_value

    def test_mape_array(self):
        y_true = np.array([1, 5, 7, 3, 4, 5])
        y_pred = np.array([-1, 3, 5, 4, 4, 3])

        mape_value = np.round(mape(y_true, y_pred), decimals=2)
        expected_mape = 56.98

        assert expected_mape == mape_value

    def test_mape_dataframe(self):
        y_true = pd.DataFrame([1, 5, 7, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 3, 5, 4, 4, 3])

        mape_value = np.round(mape(y_true, y_pred), decimals=2)
        expected_mape = 56.98

        assert expected_mape == mape_value

    def test_y_true_zero_and_ratio_zero(self):
        # If y_true = 0 and (y_pred - y_true) = 0
        y_true = [0, 1, 2, 3, 4, 5]
        y_pred = [0, 2.3, 0.4, 3.9, 3.1, 4.6]

        with pytest.raises(ValueError):
            mape(y_true, y_pred)

    def test_y_true_contains_zero(self):
        # If y_true = 0
        y_true = [0, 2, 2, 3, 4, 5]
        y_pred = [2, 2.3, 0.4, 3.9, 3.1, 4.6]

        mape_value = np.round(mape(y_true, y_pred))
        expected_mape = np.inf

        assert expected_mape == mape_value

    @given(
        arrays(
            float,
            shape=30,
            elements=floats(allow_nan=False, allow_infinity=False, min_value=1),
        ),
        arrays(
            float,
            shape=30,
            elements=floats(allow_nan=False, allow_infinity=False, min_value=1),
        ),
    )
    def test_mape_random_arrays_finite_values(self, y_true, y_pred):
        mape_value = mape(y_true, y_pred)
        expected_mape = self._correct_mape(y_true, y_pred)
        print(y_true)
        print(y_pred)

        assert expected_mape == mape_value


class TestGMAE:
    def _correct_gmae(self, y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)

        absolute_difference = np.abs(y_true - y_pred)
        return 0 if 0 in absolute_difference else gmean(absolute_difference)

    def test_wrong_vector_length(self):
        y_true = np.random.random(5)
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            gmae(y_true, y_pred)

    def test_nan_values(self):
        y_true = [np.nan, 1, 2, 3]
        y_pred = np.random.random(4)

        with pytest.raises(ValueError):
            gmae(y_true, y_pred)

    def test_infinite_values(self):
        y_true = np.random.random(4)
        y_pred = [0, np.inf, 2, 3]

        with pytest.raises(ValueError):
            gmae(y_true, y_pred)

    def test_gmae_list(self):
        y_true = [0, 1, 2, 3, 6, 5]
        y_pred = [-1, 4, 5, 10, 4, 1]

        gmae_value = np.round(gmae(y_true, y_pred), decimals=2)
        expected_gmae = 2.82

        assert expected_gmae == gmae_value

    def test_gmae_array(self):
        y_true = np.array([0, 1, 2, 3, 6, 5])
        y_pred = np.array([-1, 4, 5, 10, 4, 1])

        gmae_value = np.round(gmae(y_true, y_pred), decimals=2)
        expected_gmae = 2.82

        assert expected_gmae == gmae_value

    def test_gmae_dataframe(self):
        y_true = pd.DataFrame([0, 1, 2, 3, 6, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        gmae_value = np.round(gmae(y_true, y_pred), decimals=2)
        expected_gmae = 2.82

        assert expected_gmae == gmae_value

    def test_zero_in_difference_gmae(self):
        #if absolute difference is zero then GMAE is zero
        y_true = pd.DataFrame([0, 1, 2, 3, 4, 5])
        y_pred = pd.DataFrame([-1, 4, 5, 10, 4, 1])

        gmae_value = np.round(gmae(y_true, y_pred), decimals=2)
        expected_gmae = 0

        assert expected_gmae == gmae_value

    @given(
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
        arrays(float, shape=30, elements=floats(allow_nan=False, allow_infinity=False)),
    )
    def test_gmae_random_arrays_finite_values(self, y_true, y_pred):
        gmae_value = gmae(y_true, y_pred)
        expected_gmae = self._correct_gmae(y_true, y_pred)
        print(y_true)
        print(y_pred)
        
        assert expected_gmae == gmae_value




import numpy as np
import pandas as pd

from scipy import stats
from sklearn.base import BaseEstimator
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error


def _loglikelihood(y_pred, y_true):
    """Helper function to calculate the loglikelihood for the log likelihood chi2 test
    """

    diff = y_true - y_pred
    std_predictions = np.std(diff)
    loglikelihood = -(len(y_true) / 2) * np.log(
        2.0 * np.pi * std_predictions * std_predictions
    ) - ((np.dot(diff.T, diff)) / (2.0 * std_predictions * std_predictions))
    return loglikelihood


def pseudoinv_extended(X, ratio=1e-15):
    """Calculate pseudoinverse. Code adapted from statstools and numpy
    """

    X = np.asarray(X)
    X = X.conjugate()
    u, s, vt = np.linalg.svd(X, 0)
    s_orig = np.copy(s)
    m = u.shape[0]
    n = vt.shape[1]
    cutoff = ratio * np.maximum.reduce(s)
    for i in range(min(n, m)):
        if s[i] > cutoff:
            s[i] = 1.0 / s[i]
        else:
            s[i] = 0.0

    res = np.dot(np.transpose(vt), s[:, np.core.newaxis] * np.transpose(u))
    return res


def _ssr_f(params_dict):
    linreg_single_residues, linreg_joint_residues, max_shift, dof_joint = (
        params_dict["linreg_single_residues"],
        params_dict["linreg_joint_residues"],
        params_dict["max_shift"],
        params_dict["dof_joint"],
    )

    f_stat = (
        (linreg_single_residues - linreg_joint_residues)
        / linreg_joint_residues
        / max_shift
        * dof_joint
    )

    result_df = pd.DataFrame()
    result_df["ssr F-test"] = [
        f_stat,
        stats.f.sf(f_stat, max_shift, dof_joint),
        int(dof_joint),
        int(max_shift),
    ]
    result_df.index = ["F-value", "p-value", "degrees of freedom", "number of shifts"]
    return result_df


def _ssr_chi2(params_dict):
    data_single, linreg_single_residues, linreg_joint_residues, dof_joint, max_shift = (
        params_dict["data_single"],
        params_dict["linreg_single_residues"],
        params_dict["linreg_joint_residues"],
        params_dict["dof_joint"],
        params_dict["max_shift"],
    )

    chi2_stat = (
        len(data_single)
        * (linreg_single_residues - linreg_joint_residues)
        / linreg_joint_residues
    )

    result_df = pd.DataFrame()
    result_df["ssr_chi2test"] = [
        chi2_stat,
        stats.chi2.sf(chi2_stat, max_shift),
        int(dof_joint),
        int(max_shift),
    ]
    result_df.index = ["chi2", "p-value", "degrees of freedom", "number of shifts"]
    return result_df


def _likelihood_chi2(params_dict):
    y_pred_single, y_pred_joint, data = (
        params_dict["y_pred_single"],
        params_dict["y_pred_joint"],
        params_dict["data"],
    )
    data_single, data_joint, dof_joint = (
        params_dict["data_single"],
        params_dict["data_joint"],
        params_dict["dof_joint"],
    )
    max_shift, x_col = params_dict["max_shift"], params_dict["x_col"]

    linreg_single_loglikelihood = _loglikelihood(
        y_pred=y_pred_single, y_true=data[x_col].loc[data_single.index]
    )
    linreg_joint_loglikelihood = _loglikelihood(
        y_pred=y_pred_joint, y_true=data[x_col].loc[data_joint.index]
    )

    likelihood_ratio = -2 * (linreg_single_loglikelihood - linreg_joint_loglikelihood)

    result_df = pd.DataFrame()
    result_df["likelihood ratio test"] = [
        likelihood_ratio,
        stats.chi2.sf(likelihood_ratio, max_shift),
        int(dof_joint),
        int(max_shift),
    ]
    result_df.index = ["chi2", "p-value", "degrees of freedom", "number of shifts"]
    return result_df


def _zero_f(params_dict):
    data_joint, linreg_joint, data, y_pred_joint = (
        params_dict["data_joint"],
        params_dict["linreg_joint"],
        params_dict["data"],
        params_dict["y_pred_joint"],
    )
    linreg_joint_residues, dof_joint, max_shift, x_col = (
        params_dict["linreg_joint_residues"],
        params_dict["dof_joint"],
        params_dict["max_shift"],
        params_dict["x_col"],
    )

    constraint_matrix = np.column_stack(
        (
            np.zeros((max_shift, max_shift)),
            np.eye(max_shift, max_shift),
            np.zeros((max_shift, 1)),
        )
    )
    y_true = data[x_col].loc[data_joint.index].values
    value_restriction = np.zeros(len(constraint_matrix))

    # Parameters of the fitted linear regression model
    linreg_params = list(linreg_joint.coef_)
    linreg_params.append(linreg_joint.intercept_)
    linreg_params = np.array([linreg_params])
    constraint_params = np.dot(constraint_matrix, linreg_params.T)
    params_diff = constraint_params - value_restriction

    pseudoinv_data = pseudoinv_extended(
        np.append(data_joint.values, np.ones((len(data_joint.values), 1)), axis=1)
    )

    # Covariance matrix
    scale = linreg_joint_residues / dof_joint
    covar = np.dot(pseudoinv_data, np.transpose(pseudoinv_data)) * scale
    covar = np.dot(constraint_matrix, np.dot(covar, constraint_matrix.T))
    invcov = np.linalg.pinv(covar)

    f = (np.dot(np.dot(params_diff.T, invcov), params_diff) / len(constraint_matrix))[
        0, 0
    ]
    pvalue = stats.f.sf(f, len(constraint_matrix), dof_joint)

    result_df = pd.DataFrame()
    result_df["F-test"] = [f, pvalue, int(dof_joint), int(max_shift)]
    result_df.index = ["F-value", "p-value", "degrees of freedom", "number of shifts"]
    return result_df


STAT_TESTS = {
    "ssr_f": _ssr_f,
    "ssr_chi2": _ssr_chi2,
    "likelihood_chi2": _likelihood_chi2,
    "zero_f": _zero_f,
}


class GrangerCausality(BaseEstimator):
    """Class to check for Granger causality between two time series, i.e. 
    to check if a time series X causes Y: X->Y.

    Parameters
    ----------
    target_col : str
        The column to use as the reference, i.e. the time series Y

    x_col : str
        The column to test for Granger causality, i.e. the time 
        series X.

    max_shift : int, optional, default: 10
        The maximal number of shifts to check for Granger causality. 

    statistics : list, optional, default: ['ssr_f']
        The statistical test(s) to perform for Granger causality. A list with elements
        from the set: 'ssr_f' (sum squared residuals with F-test), 'ssr_chi2' (sum squared 
        residuals with chi square test), 'likelihood_chi2' (likelihood ratio test with 
        chi square distribution), 'zero_F' (F-test that all lag coefficients of 
        the time series X are zero).

    Attributes
    ----------
    results_ : list
        A list of pandas dataframes with the results for all the listed tests in 'statistics'.

    Examples
    --------
    >>> from gtime.causality.granger_causality import GrangerCausality
    >>> import pandas.util.testing as testing
    >>> data = testing.makeTimeDataFrame(freq="s", nper=1000)
    >>> gc = GrangerCausality(target_col='A', x_col='B', max_shift=10, statistics=['ssr_f']).fit(data)
    >>> gc.results_[0]
                        ssr F-test
    F-value               0.372640
    p-value               0.958527
    degrees of freedom  969.000000
    number of shifts     10.000000

    """

    def __init__(self, target_col: str, x_col: str, max_shift=10, statistics=["ssr_f"]):
        self.target_col = target_col
        self.x_col = x_col
        self.max_shift = max_shift
        self.statistics = statistics

    def fit(self, data: pd.DataFrame):
        """Create a dataframe with the results of the Granger causality test with the specified
        statistical test(s).

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The dataframe containing the time series.

        Returns
        -------
        self : object
            Returns the instance itself.

        """

        shifts = data.copy()
        x_columns, y_columns = [], []
        for i in range(1, self.max_shift + 1):
            shifts[f"x_shift_{i}"] = data[self.x_col].shift(i)
            shifts[f"y_shift_{i-1}"] = data[self.target_col].shift(i)
            x_columns.append(f"x_shift_{i}")
            y_columns.append(f"y_shift_{i-1}")
        shifts.drop([self.x_col, self.target_col], axis="columns", inplace=True)
        shifts = shifts.dropna()

        data_single = shifts[x_columns].copy()
        data_joint = shifts[x_columns + y_columns].copy()

        linreg_single = LinearRegression()
        linreg_joint = LinearRegression()
        y_pred_single = linreg_single.fit(
            data_single, data[self.x_col].loc[data_single.index]
        ).predict(data_single)
        y_pred_joint = linreg_joint.fit(
            data_joint, data[self.x_col].loc[data_joint.index]
        ).predict(data_joint)

        dof_single = float(data_single.shape[0] - data_single.shape[1])
        dof_joint = float(data_joint.shape[0] - data_joint.shape[1]) - 1

        linreg_single_residues = linreg_single._residues
        linreg_joint_residues = linreg_joint._residues

        self.results_ = []
        stat_test_input = {
            "linreg_single_residues": linreg_single_residues,
            "linreg_joint_residues": linreg_joint_residues,
            "dof_joint": dof_joint,
            "max_shift": self.max_shift,
            "data_single": data_single,
            "y_pred_single": y_pred_single,
            "y_pred_joint": y_pred_joint,
            "data": data,
            "x_col": self.x_col,
            "data_joint": data_joint,
            "linreg_joint": linreg_joint,
        }

        for s in self.statistics:
            self.results_.append(STAT_TESTS[s](stat_test_input))

        return self



"""
The :mod:`gtime.forecasting` module contains a collection of machine learning models,
for dealing with time series data.
"""

from .gar import GAR, GARFF
from .trend_models import TrendForecaster
from .online import HedgeForecaster

__all__ = [
    "GAR",
    "GARFF",
    "TrendForecaster",
    "HedgeForecaster",
]



import importlib
from datetime import datetime

import workalendar
from typing import Optional, Union, List

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from gtime.base import FeatureMixin, add_class_name


def _check_index(time_series: pd.DataFrame) -> None:
    index_series = pd.Series(time_series.index)
    unique_differences = pd.unique(index_series.diff().dropna().values)

    if len(unique_differences) != 1:
        raise ValueError(
            "The time series should be evenly spaced in order to obtain "
            "meaningful results."
        )

    if not isinstance(time_series.index, pd.PeriodIndex):
        raise ValueError(
            "The input time series should have a index of type "
            f"PeriodIndex, got {type(time_series.index)} instead."
        )


class Calendar(BaseEstimator, TransformerMixin, FeatureMixin):
    """Create a feature based on the national holidays of a specific country.

    The interface for this is based on the one of 'workalendar'. To see which regions
    and countries are available, check the 'workalendar'
    `documentation <https://peopledoc.github.io/workalendar/>`_.

    Parameters
    ----------
    region : str, optional, default: ``'america'``
        The region in which the ``country`` is located.

    country : str, optional, default: ``'Brazil'``
        The name of the country from which to retrieve the holidays. The country must be
        located in the given ``region``. For certain countries workalendar provides
        additional 'subregions'. In order to use them instead of the whole country,
        just use the name of the subregion instead of the country name (e.g. 'Vaud'
        instead of 'Switzerland' for the canton of Vaud which is a part of Switzerland).

    start_date : str, optional, default: ``'01/01/2019'``
        The date starting from which to retrieve the holidays.

    end_date : str, optional, default: ``'01/01/2020'``
        The date until which to retrieve the holidays.

    kernel : array-like, optional, default: ``None``
        The kernel to use when creating the feature. The holiday feature is created by
        taking the dot product between the kernel and the column which contains a 1 if
        the corresponding day is a holiday and a 0 if the day is not a holiday. The
        rolling window has the same size as the kernel and the calculated value of the
        dot product is divided by the number of holidays in the window to get the value
        of the holiday feature.

    reindex_method : str, optional, default: ``pad``
        Used only if X is passed in the ``transform`` method. It is used as the method
        with which to reindex the holiday events with the index of X. This method
        should be compatible with the reindex methods provided by pandas. Please refer
        to the pandas `documentation
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.reindex.html>`_
        for further details.

    Examples
    --------
    >>> import pandas as pd
    >>> from gtime.feature_extraction import Calendar
    >>> X = pd.DataFrame(range(0, 10), index=pd.period_range(start='2019-04-18',
    ...                  end='2019-04-27', freq='d'))
    >>> cal_feature = Calendar(region="europe", country="Italy", kernel=[2, 1])
    >>> cal_feature.fit_transform(X)
                status__Calendar
    2019-04-18               0.0
    2019-04-19               0.0
    2019-04-20               0.0
    2019-04-21               1.0
    2019-04-22               2.0
    2019-04-23               0.0
    2019-04-24               1.0
    2019-04-25               2.0
    2019-04-26               0.0
    2019-04-27               0.0

    """

    def __init__(
        self,
        region: str = "america",
        country: str = "Brazil",
        start_date: str = "01/01/2018",
        end_date: str = "01/01/2020",
        kernel: Union[List, np.ndarray] = None,
        reindex_method: str = "pad",
    ):
        super().__init__()
        self.region = region
        self.country = country
        self.start_date = start_date
        self.end_date = end_date
        self.reindex_method = reindex_method

        if kernel is None or len(kernel) == 0 or not np.isfinite(kernel).all():
            raise ValueError(
                "The kernel should be an array-like object, with at least 1 element "
                f"and should only contains finite values, got {kernel} instead."
            )
        self.kernel = kernel

    def get_feature_names(self):
        """Return feature names for output features.

        Returns
        -------
        output_feature_names : ndarray, shape (n_output_features,)
            Array of feature names.

        """

        return [self.__class__.__name__]

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the estimator. Just used to be compatible with the sklearn API.

        Parameters
        ----------
        X : pd.DataFrame, shape (n_samples, n_features)
            Input data.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        self.columns_ = X.columns.values
        return self

    @add_class_name
    def transform(self, time_series: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """Generate a DataFrame containing the events associated to the holidays of the
        selected ``country``.

        Parameters
        ----------
        time_series : pd.DataFrame, shape (n_samples, 1), optional, default: ``None``
            If provided, both ``start_date`` and ``end_date`` are going to be
            overwritten with the start and end date of the index of ``time_series``.
            Also, if provided the output DataFrame is going to be re-indexed with the
            index of ``time_series``, using the chosen ``reindex_method``.

        Returns
        -------
        events : pd.DataFrame, shape (length, 1)
            A DataFrame containing the events.

        """
        check_is_fitted(self)

        if time_series is not None:
            _check_index(time_series)

        self._initialize_start_end_date(time_series)

        workalendar_region = importlib.import_module(f".{self.region}", "workalendar")
        workalendar_country = getattr(workalendar_region, self.country)()

        events = self._get_holiday_events(workalendar_country)
        if self.kernel is not None:
            events = self._apply_kernel(events)

        aligned_events = self._align_event_indices(time_series, events)

        return aligned_events

    def _initialize_start_end_date(self, X: pd.DataFrame):
        if X is not None:
            self.start_ = X.index.values[0]
            self.end_ = X.index.values[-1]
        else:
            self.start_ = pd.Timestamp(self.start_date)
            self.end_ = pd.Timestamp(self.end_date)

        slack_period = len(self.kernel)
        slack_days = np.timedelta64(slack_period, "D")

        slacked_start_date_ = self.start_ - slack_days
        slacked_end_date_ = self.end_ + slack_days

        self.slacked_start_date_ = datetime(
            year=slacked_start_date_.year,
            month=slacked_start_date_.month,
            day=slacked_start_date_.day,
        )
        self.slacked_end_date_ = datetime(
            year=slacked_end_date_.year,
            month=slacked_end_date_.month,
            day=slacked_end_date_.day,
        )

    def _get_holiday_events(self, country_mod):
        index = pd.date_range(
            start=self.slacked_start_date_, end=self.slacked_end_date_, freq="D"
        )
        years = index.year.unique()
        events = pd.DataFrame()
        for year in years:
            events = events.append(country_mod.holidays(year))

        events = events.rename(columns={0: "date", 1: "events"})
        events["date"] = pd.to_datetime(events["date"])
        events["status"] = 1
        events = events.set_index("date")
        events = self._group_by_event_name(events)

        events = events.reindex(index)
        events = events.drop(columns=["events"])

        events["status"] = events["status"].fillna(0).astype(int)
        events = events.sort_index()

        return events

    def _group_by_event_name(self, events_df):
        def aggregate_events_same_day(event):
            return pd.Series(
                {"status": 1, "events": " & ".join(event["events"].tolist())}
            )

        grouped_events = events_df.groupby(events_df.index).apply(
            aggregate_events_same_day
        )
        return grouped_events

    def _apply_kernel(self, events):
        kernel_events = events.copy()
        klen = len(self.kernel)

        def apply_kernel(w):
            if sum(w) != 0:
                return np.dot(w, self.kernel) / sum(w)
            else:
                return 0

        kernel_events["status"] = (
            kernel_events["status"]
            .rolling(klen, center=True)
            .apply(apply_kernel, raw=False)
        )
        return kernel_events

    def _align_event_indices(self, ts, events):
        if ts is not None:
            X = ts.copy()
            new_x_line = pd.DataFrame(
                columns=X.columns, index=pd.PeriodIndex([X.index[-1] + 1])
            )
            X_to_cut = pd.concat([X, new_x_line], axis=0)
            bins = pd.cut(events.index, X_to_cut.index.to_timestamp())

            grouped_events = events.groupby(bins).mean().ffill()

            grouped_events.index = pd.to_datetime(
                grouped_events.index.map(lambda row_index: row_index.left)
            )
            X.index = X.index.to_timestamp()

            # events_renamed = pd.merge(X, grouped_events, left_index=True, right_index=True, how="inner")

            # filtering by index instead of merge
            events_renamed = grouped_events.loc[X.index, :]
            events_renamed.index = ts.index

        else:
            events_renamed = events[self.start_ : self.end_]

        return events_renamed



import functools

from sklearn.utils.validation import check_is_fitted


def add_class_name(func):
    @functools.wraps(func)
    def wrapper_add_class_name(*args, **kwargs):
        value = func(*args, **kwargs)
        return value.add_suffix("__" + args[0].__class__.__name__)

    return wrapper_add_class_name


class FeatureMixin:
    """Mixin class for all feature extraction estimators in giotto-time."""

    _estimator_type = "feature_extractor"

    def get_feature_names(self):
        """Return feature names for output features.

        Returns
        -------
        output_feature_names : ndarray, shape (n_output_features,)
            Array of feature names.

        """
        check_is_fitted(self)

        return [f"{name}__{self.__class__.__name__}" for name in self.columns_]



import pandas as pd
import numpy as np
from gtime.forecasting.online import HedgeForecaster


def test_hedge_fit_predict():
    time_index = pd.date_range("2020-01-01", "2020-01-20")
    X_np = np.concatenate(
        (np.random.randint(4, size=(20, 2)), np.array([100] * 20).reshape(-1, 1)),
        axis=1,
    )
    X = pd.DataFrame(X_np, index=time_index)
    y = pd.DataFrame(
        np.random.randint(4, size=(20, 1)), index=time_index, columns=["y_1"]
    )
    hr = HedgeForecaster(random_state=42)

    preds = hr.fit_predict(X, y)
    np.testing.assert_equal(preds.shape, y.shape)
    np.testing.assert_almost_equal(hr.weights_[0], hr.weights_[1], decimal=2)
    assert hr.weights_[2] < hr.weights_[0]
    assert hr.weights_[2] < hr.weights_[1]



import pandas as pd
from sklearn.base import TransformerMixin, BaseEstimator

from gtime.causality.base import CausalityMixin


class ShiftedPearsonCorrelation(BaseEstimator, TransformerMixin, CausalityMixin):
    """Class responsible for assessing the shifted Pearson correlations (PPMCC) between
    two or more series. For more info about the test, click
    `here <https://statistics.laerd.com/statistical-guides/pearson-correlation-coefficient-statistical-guide.php>`_.

    Parameters
    ----------
    min_shift : int, optional, default: ``1``
        The minimum number of shifts to check for.

    max_shift : int, optional, default: ``10``
        The maximum number of shifts to check for.

    target_col : str, optional, default: ``None``
            The column to use as the a reference (i.e., the columns which is not
            shifted).

    dropna : bool, optional, default: ``False``
        Determines if the Nan values created by shifting are retained or dropped.

    bootstrap_iterations : int, optional, default: ``None``
        If not None, compute the p_values of the test, by performing bootstrapping of
        the original data (sampling with replacement).

    permutation_iterations : int, optional, default: ``None``
        If not None, compute the p_values of the test, by performing permutations of
        the original data.

    Examples
    --------
    >>> from gtime.causality.pearson_correlation import ShiftedPearsonCorrelation
    >>> import pandas.util.testing as testing
    >>> data = testing.makeTimeDataFrame(freq="s")
    >>> spc = ShiftedPearsonCorrelation(target_col="A")
    >>> spc.fit(data)
    >>> spc.best_shifts_
    y  A  B  C  D
    x
    A  8  9  6  5
    B  7  4  4  6
    C  3  4  9  9
    D  7  1  9  1
    >>> spc.max_corrs_
    y         A         B         C         D
    x
    A  0.383800  0.260627  0.343628  0.360151
    B  0.311608  0.307203  0.255969  0.298523
    C  0.373613  0.267335  0.211913  0.140034
    D  0.496535  0.204770  0.402473  0.310065
    """

    def __init__(
        self,
        min_shift: int = 1,
        max_shift: int = 10,
        target_col: str = None,
        dropna: bool = False,
        bootstrap_iterations: int = None,
        permutation_iterations: int = None,
    ):
        super().__init__(
            bootstrap_iterations=bootstrap_iterations,
            permutation_iterations=permutation_iterations,
        )
        self.min_shift = min_shift
        self.max_shift = max_shift
        self.target_col = target_col
        self.dropna = dropna

    def fit(self, data: pd.DataFrame) -> "ShiftedPearsonCorrelation":
        """Create the dataframe of shifts of each time series which maximize the
         Pearson correlation (PPMCC).

        Parameters
        ----------
        data : pd.DataFrame, shape (n_samples, n_time_series), required
            The DataFrame containing the time series on which to compute the shifted
            correlations.

        Returns
        -------
        self : ``ShiftedPearsonCorrelation``

        """
        best_shifts = self._compute_best_shifts(data, self._get_max_corr_shift)

        pivot_tables = self._create_pivot_tables(best_shifts)

        self.best_shifts_ = pivot_tables["best_shifts"]
        self.max_corrs_ = pivot_tables["max_corrs"]

        if self.bootstrap_iterations:
            self.bootstrap_p_values_ = pivot_tables["bootstrap_p_values"]

        if self.permutation_iterations:
            self.permutation_p_values_ = pivot_tables["permutation_p_values"]

        return self

    def _get_max_corr_shift(self, data: pd.DataFrame, x, y):
        shifts = pd.DataFrame()

        for shift in range(self.min_shift, self.max_shift + 1):
            shifts[shift] = data[x].shift(shift)

        shifts = shifts.dropna()
        self.shifted_corrs = shifts.corrwith(data[y])

        q = self.shifted_corrs.max(), self.shifted_corrs.idxmax()
        return q



"""
The :mod:`gtime.preprocessing` module deals with the preprocessing of time series
data.
"""

from .time_series_conversion import (
    _SequenceToTimeIndexSeries,
    _PandasSeriesToTimeIndexSeries,
    _TimeIndexSeriesToPeriodIndexSeries,
)

from .time_series_preparation import TimeSeriesPreparation

__all__ = [
    "TimeSeriesPreparation",
]



import pandas as pd
import pytest
from hypothesis import given, strategies as st
if pd.__version__ >= '1.0.0':
    import pandas._testing as testing
else:
    import pandas.util.testing as testing

from gtime.feature_generation import Constant, PeriodicSeasonal
from gtime.utils.hypothesis.time_indexes import giotto_time_series


class TestPeriodicSesonalFeature:
    def test_missing_start_date_or_period(self):
        periodic_feature = PeriodicSeasonal()
        with pytest.raises(ValueError):
            periodic_feature.transform()

        periodic_feature = PeriodicSeasonal(index_period=1)
        with pytest.raises(ValueError):
            periodic_feature.transform()

        periodic_feature = PeriodicSeasonal(start_date="2010-01-01")
        with pytest.raises(ValueError):
            periodic_feature.transform()

    def test_string_period(self):
        testing.N, testing.K = 20, 1
        ts = testing.makeTimeDataFrame(freq="s")
        periodic_feature = PeriodicSeasonal(period="1 days")
        periodic_feature.transform(ts)

        assert type(periodic_feature.period) == pd.Timedelta

    def test_correct_start_date(self):
        testing.N, testing.K = 20, 1
        ts = testing.makeTimeDataFrame(freq="s")
        start_date = "2018-01-01"
        periodic_feature = PeriodicSeasonal(period="1 days", start_date=start_date)
        periodic_feature.transform(ts)

        assert periodic_feature.start_date == ts.index.values[0]

        periodic_feature = PeriodicSeasonal(
            period="3 days", index_period=10, start_date=start_date
        )
        periodic_feature.transform()
        assert periodic_feature.start_date == pd.to_datetime(start_date)

        start_date = pd.to_datetime("2018-01-01")
        periodic_feature = PeriodicSeasonal(
            period="3 days", index_period=10, start_date=start_date
        )
        periodic_feature.transform()
        assert periodic_feature.start_date == start_date

    def test_too_high_sampling_frequency(self):
        start_date = "2018-01-01"
        periodic_feature = PeriodicSeasonal(
            period="2 days",
            start_date=start_date,
            index_period=pd.date_range(start=start_date, end="2020-01-01", freq="W"),
        )
        with pytest.raises(ValueError):
            periodic_feature.transform()

    def test_correct_sinusoide(self):
        testing.N, testing.K = 30, 1
        ts = testing.makeTimeDataFrame(freq="MS")
        start_date = "2018-01-01"
        periodic_feature = PeriodicSeasonal(
            period="365 days",
            start_date=start_date,
            index_period=pd.date_range(start=start_date, end="2020-01-01", freq="W"),
        )
        output_sin = periodic_feature.transform(ts)
        expected_index = pd.DatetimeIndex(
            [
                "2000-01-01",
                "2000-02-01",
                "2000-03-01",
                "2000-04-01",
                "2000-05-01",
                "2000-06-01",
                "2000-07-01",
                "2000-08-01",
                "2000-09-01",
                "2000-10-01",
                "2000-11-01",
                "2000-12-01",
                "2001-01-01",
                "2001-02-01",
                "2001-03-01",
                "2001-04-01",
                "2001-05-01",
                "2001-06-01",
                "2001-07-01",
                "2001-08-01",
                "2001-09-01",
                "2001-10-01",
                "2001-11-01",
                "2001-12-01",
                "2002-01-01",
                "2002-02-01",
                "2002-03-01",
                "2002-04-01",
                "2002-05-01",
                "2002-06-01",
            ],
            dtype="datetime64[ns]",
            freq="MS",
        )
        expected_df = pd.DataFrame.from_dict(
            {
                f"0__{periodic_feature.__class__.__name__}": [
                    0.0,
                    0.25433547,
                    0.42938198,
                    0.49999537,
                    0.43585316,
                    0.25062091,
                    0.0043035,
                    -0.25062091,
                    -0.43585316,
                    -0.49999537,
                    -0.42938198,
                    -0.24688778,
                    0.00860668,
                    0.2617078,
                    0.42938198,
                    0.49999537,
                    0.43585316,
                    0.25062091,
                    0.0043035,
                    -0.25062091,
                    -0.43585316,
                    -0.49999537,
                    -0.42938198,
                    -0.24688778,
                    0.00860668,
                    0.2617078,
                    0.42938198,
                    0.49999537,
                    0.43585316,
                    0.25062091,
                ]
            }
        )
        expected_df.index = expected_index
        pd.testing.assert_frame_equal(output_sin, expected_df)


class TestConstantFeature:
    def test_correct_constant_feature(self):
        constant = 12
        df = pd.DataFrame.from_dict({"old_name": [0, 1, 2, 3, 4, 5]})

        constant_feature = Constant(constant=constant)

        df_constant = constant_feature.fit_transform(df)
        expected_df_constant = pd.DataFrame.from_dict(
            {
                f"0__{constant_feature.__class__.__name__}": [
                    constant,
                    constant,
                    constant,
                    constant,
                    constant,
                    constant,
                ]
            }
        )

        testing.assert_frame_equal(expected_df_constant, df_constant, check_dtype=False)

    @given(
        giotto_time_series(
            min_length=1,
            start_date=pd.Timestamp(2000, 1, 1),
            end_date=pd.Timestamp(2010, 1, 1),
        ),
        st.integers(0, 100),
    )
    def test_random_ts_and_constant(self, df: pd.DataFrame, constant: int):

        constant_feature = Constant(constant=constant)
        df_constant = constant_feature.fit_transform(df)

        #  testing.assert_frame_equal(expected_df_constant, df_constant)



from datetime import datetime
from typing import Tuple

import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from hypothesis import given, settings, HealthCheck
from hypothesis.extra.numpy import arrays
from pandas.testing import assert_series_equal, assert_frame_equal

from gtime.preprocessing import TimeSeriesPreparation
from gtime.utils.hypothesis.time_indexes import (
    series_with_timedelta_index,
    series_with_period_index,
    series_with_datetime_index,
    available_freqs,
)
from .utils import (
    pandas_series_with_period_index,
    datetime_index_series_to_period_index_series,
    timedelta_index_series_to_period_index_series,
)


class TestConstructor:
    @given(st.datetimes(), available_freqs(), st.booleans(), st.text())
    def test_constructor_does_not_fail(
        self,
        start: datetime,
        freq: pd.Timedelta,
        resample_if_not_equispaced: bool,
        output_name: str,
    ):
        TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=resample_if_not_equispaced,
            output_name=output_name,
        )

    @given(st.datetimes(), available_freqs(), st.booleans(), st.text())
    def test_constructor_initializes_parameters(
        self,
        start: datetime,
        freq: pd.Timedelta,
        resample_if_not_equispaced: bool,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=resample_if_not_equispaced,
            output_name=output_name,
        )

        assert time_series_preparation.start == start
        assert time_series_preparation.end is None
        assert time_series_preparation.freq == freq
        assert (
            time_series_preparation.resample_if_not_equispaced
            == resample_if_not_equispaced
        )
        assert time_series_preparation.output_name == output_name


class TestToTimeIndexSeries:
    @given(st.lists(st.floats()), st.datetimes(), available_freqs())
    def test_list_as_input(
        self, input_list: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(input_list)
        expected_time_series = pandas_series_with_period_index(
            input_list, start, freq=freq
        )
        assert_series_equal(computed_time_series, expected_time_series)

    @given(
        arrays(shape=st.integers(0, 1000), dtype=float),
        st.datetimes(),
        available_freqs(),
    )
    def test_array_as_input(
        self, input_array: np.ndarray, start: datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            input_array
        )
        expected_time_series = pandas_series_with_period_index(
            input_array, start, freq=freq
        )
        assert_series_equal(computed_time_series, expected_time_series)

    @given(series_with_period_index(), st.datetimes(), available_freqs())
    def test_period_index_series_unchanged(
        self, period_index_series: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            period_index_series
        )
        assert_series_equal(computed_time_series, period_index_series)

    @given(series_with_datetime_index(), st.datetimes(), available_freqs())
    def test_datetime_index_series_unchanged(
        self, datetime_index_series: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            datetime_index_series
        )
        assert_series_equal(computed_time_series, datetime_index_series)

    @given(series_with_timedelta_index(), st.datetimes(), available_freqs())
    def test_timedelta_index_series_unchanged(
        self, timedelta_index_series: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            timedelta_index_series
        )
        assert_series_equal(computed_time_series, timedelta_index_series)

    @given(st.tuples())
    def test_wrong_input_type(self, wrong_input: Tuple):
        time_series_preparation = TimeSeriesPreparation()
        with pytest.raises(TypeError):
            time_series_preparation._to_time_index_series(wrong_input)

    @given(series_with_period_index(), st.datetimes(), available_freqs())
    def test_period_index_dataframe_unchanged(
        self, period_index_series: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        period_index_dataframe = pd.DataFrame(period_index_series)
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            period_index_dataframe
        )
        assert_series_equal(computed_time_series, period_index_series)

    @given(series_with_datetime_index(), st.datetimes(), available_freqs())
    def test_datetime_index_dataframe_unchanged(
        self, datetime_index_series: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        datetime_index_dataframe = pd.DataFrame(datetime_index_series)
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            datetime_index_dataframe
        )
        assert_series_equal(computed_time_series, datetime_index_series)

    @given(series_with_timedelta_index(), st.datetimes(), available_freqs())
    def test_timedelta_index_dataframe_unchanged(
        self, timedelta_index_series: pd.Series, start: datetime, freq: pd.Timedelta,
    ):
        timedelta_index_dataframe = pd.DataFrame(timedelta_index_series)
        time_series_preparation = TimeSeriesPreparation(start=start, freq=freq)
        computed_time_series = time_series_preparation._to_time_index_series(
            timedelta_index_dataframe
        )
        assert_series_equal(computed_time_series, timedelta_index_series)


class TestToEquispacedTimeSeries:
    @given(
        series_with_period_index(), st.datetimes(), available_freqs(), st.text(),
    )
    def test_with_resample_false(
        self, series: pd.Series, start: datetime, freq: pd.Timedelta, output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=False,
            output_name=output_name,
        )
        computed_series = time_series_preparation._to_equispaced_time_series(series)
        assert_series_equal(computed_series, series)

    @given(series_with_period_index(), st.datetimes(), available_freqs(), st.text())
    def test_with_resample_true(
        self, series: pd.Series, start: datetime, freq: pd.Timedelta, output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start,
            freq=freq,
            resample_if_not_equispaced=True,
            output_name=output_name,
        )
        with pytest.raises(NotImplementedError):
            time_series_preparation._to_equispaced_time_series(series)


class TestToPeriodIndex:
    @settings(suppress_health_check=[HealthCheck.filter_too_much])
    @given(series_with_period_index())
    def test_with_period_index_as_input(self, period_index_series):
        time_series_preparation = TimeSeriesPreparation()
        computed_time_series = time_series_preparation._to_period_index_time_series(
            period_index_series
        )
        assert_series_equal(computed_time_series, period_index_series)

    @given(series_with_datetime_index())
    def test_with_datetime_index_as_input(self, datetime_index_series):
        time_series_preparation = TimeSeriesPreparation()
        computed_time_series = time_series_preparation._to_period_index_time_series(
            datetime_index_series
        )
        expected_time_series = datetime_index_series_to_period_index_series(
            computed_time_series
        )
        assert_series_equal(computed_time_series, expected_time_series)

    @given(series_with_timedelta_index())
    def test_with_timedelta_index_as_input(self, timedelta_index_series):
        time_series_preparation = TimeSeriesPreparation()
        computed_time_series = time_series_preparation._to_period_index_time_series(
            timedelta_index_series
        )
        expected_time_series = timedelta_index_series_to_period_index_series(
            timedelta_index_series
        )
        assert_series_equal(computed_time_series, expected_time_series)


class TestToPeriodIndexDataFrame:
    @given(series_with_period_index(), st.text())
    def test_output_dataframe_is_correct(
        self, period_index_series: pd.Series, output_name: str
    ):
        time_series_preparation = TimeSeriesPreparation(output_name=output_name)
        computed_time_series = time_series_preparation._to_period_index_dataframe(
            period_index_series
        )
        expected_time_series = pd.DataFrame({output_name: period_index_series})
        assert_frame_equal(computed_time_series, expected_time_series)


class TestTransform:
    @given(st.lists(st.floats()), st.datetimes(), available_freqs(), st.text())
    def test_list_as_input(
        self,
        input_list: pd.Series,
        start: datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(input_list)
        expected_series = pandas_series_with_period_index(input_list, start, freq=freq)
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(
        arrays(shape=st.integers(0, 1000), dtype=float),
        st.datetimes(),
        available_freqs(),
        st.text(),
    )
    def test_array_as_input(
        self,
        input_array: np.ndarray,
        start: datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(input_array)
        expected_series = pandas_series_with_period_index(input_array, start, freq=freq)
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(series_with_period_index(), st.datetimes(), available_freqs(), st.text())
    def test_period_index_as_input(
        self,
        period_index_series: pd.Series,
        start: datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(period_index_series)
        expected_time_series = pd.DataFrame({output_name: period_index_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(series_with_datetime_index(), st.datetimes(), available_freqs(), st.text())
    def test_datetime_index_as_input(
        self,
        datetime_index_series: pd.Series,
        start: datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(datetime_index_series)
        expected_series = datetime_index_series_to_period_index_series(
            datetime_index_series, freq=freq
        )
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)

    @given(series_with_timedelta_index(), st.datetimes(), available_freqs(), st.text())
    def test_timedelta_index_as_input(
        self,
        timedelta_index_series: pd.Series,
        start: datetime,
        freq: pd.Timedelta,
        output_name: str,
    ):
        time_series_preparation = TimeSeriesPreparation(
            start=start, freq=freq, output_name=output_name
        )
        computed_time_series = time_series_preparation.transform(timedelta_index_series)
        expected_series = timedelta_index_series_to_period_index_series(
            timedelta_index_series, freq=freq
        )
        expected_time_series = pd.DataFrame({output_name: expected_series})
        assert_frame_equal(computed_time_series, expected_time_series)



from typing import Tuple

import numpy as np
import pandas as pd
from hypothesis import given
from sklearn.compose import make_column_selector

from gtime.compose import FeatureCreation
from gtime.feature_extraction import Shift, MovingAverage
from gtime.utils.hypothesis.feature_matrices import X_y_matrices, X_matrices, y_matrices

df_transformer = FeatureCreation(
    [
        ("shift_0", Shift(0), make_column_selector(dtype_include=np.number)),
        ("shift_1", Shift(1), make_column_selector(dtype_include=np.number)),
        (
            "moving_average_3",
            MovingAverage(window_size=3),
            make_column_selector(dtype_include=np.number),
        ),
    ]
)


class TestXyMatrices:
    @given(X_y_matrices(horizon=3, df_transformer=df_transformer))
    def test_X_shape_correct(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert X.shape[1] == len(df_transformer.transformers_)

    @given(X_y_matrices(horizon=3, df_transformer=df_transformer))
    def test_shape_consistent(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert X.shape[0] == y.shape[0]

    @given(
        X_y_matrices(horizon=3, df_transformer=df_transformer, allow_nan_infinity=False)
    )
    def test_allow_nan_false(self, X_y: Tuple[pd.DataFrame, pd.DataFrame]):
        X, y = X_y
        assert X.dropna(axis=0, how="any").shape[0] == max(X.shape[0] - 2, 0)
        assert y.dropna(axis=0, how="any").shape[0] == max(y.shape[0] - 3, 0)


class TestXMatrices:
    @given(X_matrices(df_transformer=df_transformer))
    def test_X_shape_correct(self, X: pd.DataFrame):
        assert X.shape[1] == len(df_transformer.transformers_)

    @given(X_matrices(df_transformer=df_transformer))
    def test_X_column_names(self, X: pd.DataFrame):
        assert list(X.columns) == [
            "shift_0__time_series__Shift",
            "shift_1__time_series__Shift",
            "moving_average_3__time_series__MovingAverage",
        ]

    @given(X_matrices(df_transformer=df_transformer, allow_nan_infinity=False))
    def test_allow_nan_false(self, X: pd.DataFrame):
        assert X.dropna(axis=0, how="any").shape[0] == max(X.shape[0] - 2, 0)


class TestYMatrices:
    @given(y_matrices(horizon=1))
    def test_horizon_1(self, y: pd.DataFrame):
        assert y.shape[1] == 1

    @given(y_matrices(horizon=2))
    def test_horizon_2(self, y: pd.DataFrame):
        assert y.shape[1] == 2

    @given(y_matrices(horizon=3))
    def test_horizon_3(self, y: pd.DataFrame):
        assert y.shape[1] == 3

    @given(y_matrices(horizon=3))
    def test_y_column_names(self, y: pd.DataFrame):
        assert list(y.columns) == ["y_1", "y_2", "y_3"]

    @given(y_matrices(horizon=3, allow_nan_infinity=False))
    def test_allow_nan_false(self, y: pd.DataFrame):
        assert y.dropna(axis=0, how="any").shape[0] == max(y.shape[0] - 3, 0)



from datetime import datetime
from typing import Optional, Dict, Union, Iterable

import hypothesis.strategies as st
import numpy as np
import pandas as pd
from hypothesis import assume
from hypothesis.extra.numpy import arrays

from .utils import (
    freq_to_timedelta,
    initialize_start_date_end_date,
    order_pair,
    initialize_start_timedelta_end_timedelta,
    expected_index_length_from,
    expected_start_date_from,
    expected_end_date_from,
)

IndexRangeArgs = Dict[str, Union[datetime, int, pd.Timedelta]]

string_freqs = ["B", "D", "W", "M", "Q", "A", "Y", "H", "T", "S"]
pandas_range_params = ["start", "end", "periods", "freq"]


@st.composite
def giotto_time_series(
    draw,
    start_date: Optional[pd.Timestamp] = None,
    end_date: Optional[pd.Timestamp] = None,
    min_length: int = 0,
    max_length: int = 1000,
    name: str = "time_series",
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    period_index_series = draw(
        series_with_period_index(
            start=start_date,
            end=end_date,
            min_length=min_length,
            max_length=max_length,
            allow_nan=allow_nan,
            allow_infinity=allow_infinity,
        )
    )
    return pd.DataFrame({name: period_index_series})


@st.composite
def series_with_period_index(
    draw,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    """ Returns a strategy to generate a Pandas Series with PeriodIndex

    Parameters
    ----------
    draw
    start : ``datetime``, optional, (default=None)
    end : ``datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=None)
    allow_nan : ``bool``, optional, (default=True)
    allow_infinity : ``bool``, optional, (default=True)

    Returns
    -------
    pd.Series with PeriodIndex
    """
    index = draw(
        period_indexes(
            start=start, end=end, min_length=min_length, max_length=max_length
        )
    )
    values = draw(
        arrays(
            dtype=np.float64,
            shape=index.shape[0],
            elements=st.floats(
                allow_nan=allow_nan, allow_infinity=allow_infinity, width=32
            ),
        )
    )
    return pd.Series(index=index, data=values)


@st.composite
def series_with_datetime_index(
    draw,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    """ Returns a strategy to generate a Pandas Series with DatetimeIndex

    Parameters
    ----------
    draw
    start : ``datetime``, optional, (default=None)
    end : ``datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    allow_nan : ``bool``, optional, (default=True)
    allow_infinity : ``bool``, optional, (default=True)

    Returns
    -------
    pd.Series with DatetimeIndex
    """
    index = draw(
        datetime_indexes(
            start=start, end=end, min_length=min_length, max_length=max_length
        )
    )
    values = draw(
        arrays(
            dtype=np.float64,
            shape=index.shape[0],
            elements=st.floats(
                allow_nan=allow_nan, allow_infinity=allow_infinity, width=32
            ),
        )
    )
    return pd.Series(index=index, data=values)


@st.composite
def series_with_timedelta_index(
    draw,
    start: Optional[pd.Timedelta] = None,
    end: Optional[pd.Timedelta] = None,
    min_length: int = 0,
    max_length: int = 1000,
    allow_nan: bool = True,
    allow_infinity: bool = True,
):
    """ Returns a strategy to generate a Pandas Series with Timedelta index

    Parameters
    ----------
    draw
    start : ``pd.Timedelta``, optional, (default=None)
    end : ``pd.Timedelta``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=None)
    allow_nan : ``bool``, optional, (default=True)
    allow_infinity : ``bool``, optional, (default=True)

    Returns
    -------
    pd.Series with TimedeltaIndex
    """
    index = draw(
        timedelta_indexes(
            start=start, end=end, min_length=min_length, max_length=max_length
        )
    )
    values = draw(
        arrays(
            dtype=np.float64,
            shape=index.shape[0],
            elements=st.floats(
                allow_nan=allow_nan, allow_infinity=allow_infinity, width=32
            ),
        )
    )
    return pd.Series(index=index, data=values)


@st.composite
def period_indexes(
    draw,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
):
    """ Returns a strategy to generate Pandas PeriodIndex

    Parameters
    ----------
    draw
    start : ``datetime``, optional, (default=None)
    end : ``datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)

    Returns
    -------
    LazyStrategy that generates pandas PeriodIndex
    """
    period_range_args = draw(
        period_range_args_from(
            start=start, end=end, min_length=min_length, max_length=max_length
        )
    )
    assume(
        period_range_args_are_correct(
            period_range_args, min_length, max_length, start, end
        )
    )
    return compute_period_range(period_range_args)


@st.composite
def datetime_indexes(
    draw,
    start: Optional[datetime] = None,
    end: Optional[datetime] = None,
    min_length: int = 0,
    max_length: int = 1000,
):
    """ Returns a strategy to generate Pandas DatetimeIndex.

    Parameters
    ----------
    draw
    start : ``datetime``, optional, (default=None)
    end : ``datetime``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, default: ``1000``

    Returns
    -------
    LazyStrategy that generates pd.DatetimeIndex
    """
    date_range_args = draw(date_range_args_from(start, end, min_length, max_length))
    assume(
        date_range_args_are_correct(date_range_args, min_length, max_length, start, end)
    )
    return compute_date_range(date_range_args)


@st.composite
def timedelta_indexes(
    draw,
    start: Optional[pd.Timedelta] = None,
    end: Optional[pd.Timedelta] = None,
    min_length: int = 0,
    max_length: int = 1000,
):
    """ Returns a strategy to generate Pandas TimedeltaIndex

    Parameters
    ----------
    draw
    start : ``pd.Timedelta``, optional, (default=None)
    end : ``pd.Timedelta``, optional, (default=None)
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, (default=1000)

    Returns
    -------
    LazyStrategy that generates pd.TimedeltaIndex
    """
    timedelta_range_args = draw(
        timedelta_range_args_from(start, end, min_length, max_length)
    )
    assume(
        timedelta_range_args_are_correct(
            timedelta_range_args,
            min_length=min_length,
            max_length=max_length,
            min_start=start,
            max_end=end,
        )
    )
    return compute_timedelta_range(timedelta_range_args)


def pair_of_ordered_dates(
    start: Optional[datetime] = None, end: Optional[datetime] = None
):
    """ Returns an hypothesis strategy that generates a pair of ordered
    datetime. Useful to create a Pandas index

    Parameters
    ----------
    start : ``datetime``, optional, (default=None)
    end : ``datetime``, optional, (default=None)

    Returns
    -------
    LazyStrategy that generates Tuple[datetime, datetime]
    """
    start, end = initialize_start_date_end_date(start, end)

    date1 = st.datetimes(start, end)
    date2 = st.datetimes(start, end)

    return order_pair(date1, date2)


def pair_of_ordered_timedeltas(
    start: Optional[pd.Timedelta] = None, end: Optional[pd.Timedelta] = None
):
    """ Returns an hypothesis strategy that generates a pair of ordered
    pd.Timedelta. Useful to create a Pandas TimedeltaIndex

    Parameters
    ----------
    start : ``datetime``, optional, (default=None)
    end : ``datetime``, optional, (default=None)

    Returns
    -------
    LazyStrategy that generates Tuple[pd.Timedelta, pd.Timedelta]
    """
    start, end = initialize_start_timedelta_end_timedelta(start, end)

    timedelta1 = st.timedeltas(start, end)
    timedelta2 = st.timedeltas(start, end)

    return order_pair(timedelta1, timedelta2)


def positive_bounded_integers(max_value):
    """ Returns a strategy that generates a positive bounded integer

    Parameters
    ----------
    max_value : ``int``, required
        maximum value of the integer

    Returns
    -------
    LazyStrategy
    """
    return st.integers(min_value=0, max_value=max_value)


def samples_from(iterable: Iterable):
    """ Strategy that samples from an iterable.

    Parameters
    ----------
    iterable : ``Iterable[Generic]``, required

    Returns
    -------
    LazyStrategy
    """
    return st.builds(lambda index: iterable[index], st.integers(0, len(iterable) - 1))


@st.composite
def available_freqs(draw) -> pd.Timedelta:
    """ Strategy that samples from the available frequencies and returns a
    corresponding pd.Timedelta

    Parameters
    ----------
    draw

    Returns
    -------
    pd.Timedelta
    """
    return freq_to_timedelta(draw(samples_from(string_freqs)))


@st.composite
def period_range_args_from(
    draw,
    start: datetime,
    end: datetime,
    min_length: int,
    max_length: int,
    arg_to_remove: Optional[str] = None,
) -> IndexRangeArgs:
    """ Strategy that computes arguments to call pd.period_range() from
    start, end and max length of the resulting period
    It returns a dictionary with 3 out of the 3 keys 'start', 'end', 'periods'
    and 'freq'. One of them is automatically removed since pd.period_range()
    accepts only 3 of the 4 inputs.

    Parameters
    ----------
    draw
    start : ``datetime``, required
    end : ``datetime``, required
    min_length : ``int``, required
    max_length : ``int``, required
    arg_to_remove : ``str``, optional, (default=None)

    Returns
    -------
    IndexRangeArgs
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.
    """
    start, end = draw(pair_of_ordered_dates(start, end))
    periods = draw(st.integers(min_value=min_length, max_value=max_length))
    freq = draw(available_freqs())
    element_to_exclude = (
        draw(samples_from(pandas_range_params))
        if arg_to_remove is None
        else arg_to_remove
    )
    assume(element_to_exclude != "freq")

    args = {"start": start, "end": end, "periods": periods, "freq": freq}
    del args[element_to_exclude]

    return args


def period_range_args_are_correct(
    period_range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: Optional[pd.Period] = None,
    max_end: Optional[pd.Period] = None,
) -> bool:
    """ Returns True if the period range arguments are correct. i.e. if
    'periods' is not a key it checks that the length of the resulting period
    index is below ``max_length``.

    Parameters
    ----------
    period_range_args : ``IndexRangeArgs``, required
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.

    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``pd.Period``, optional, (default=1000)
    max_end : ``pd.Period``, optional, (default=1000


    Returns
    -------
    bool
    """
    min_start = min_start if min_start is not None else pd.Timestamp("1980-01-01")
    max_end = max_end if max_end is not None else pd.Timestamp("2020-01-01")
    return _range_args_are_correct(
        period_range_args, min_length, max_length, min_start, max_end
    )


def compute_period_range(period_range_args: IndexRangeArgs) -> pd.PeriodIndex:
    """ Returns a Pandas PeriodIndex from the given ``period_range_args``.

    Parameters
    ----------
    period_range_args : ``IndexRangeArgs``, required

    Returns
    -------
    pd.PeriodIndex
    """
    return pd.period_range(**period_range_args)


@st.composite
def date_range_args_from(
    draw,
    start: datetime,
    end: datetime,
    min_length: int,
    max_length: int,
    arg_to_remove: Optional[str] = None,
) -> IndexRangeArgs:
    """ Strategy that computes arguments to call pd.date_range() from
    start, end and max length of the resulting period
    It returns a dictionary with 3 out of the 3 keys 'start', 'end', 'periods'
    and 'freq'. One of them is automatically removed since pd.date_range()
    accepts only 3 of the 4 inputs.

    Parameters
    ----------
    draw
    start : ``datetime``, required
    end : ``datetime``, required
    min_length: ``int``, required
    max_length : ``int``, required
    arg_to_remove: ``str``, optional, (default=None)

    Returns
    -------
    IndexRangeArgs
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.
    """
    start, end = draw(pair_of_ordered_dates(start, end))
    periods = draw(st.integers(min_value=min_length, max_value=max_length))
    freq = draw(available_freqs())
    element_to_exclude = (
        draw(samples_from(pandas_range_params))
        if arg_to_remove is None
        else arg_to_remove
    )

    args = {"start": start, "end": end, "periods": periods, "freq": freq}
    del args[element_to_exclude]

    return args


def date_range_args_are_correct(
    date_range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: pd.Timestamp = None,
    max_end: pd.Timestamp = None,
) -> bool:
    """ Returns True if the date range arguments are correct.
    It checks 3 things:

    - if 'periods' is not a key it checks that the length of the resulting
    period index is below ``max_length``.
    - if 'start' is not a keyword it checks that the expected start date is
    after ``min_start``
    - if 'end' is not a keyword it checks that the expected end date is
    before ``max_end``

    Parameters
    ----------
    date_range_args : ``IndexRangeArgs``, required
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.

    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``pd.Timestamp``, optional, (default=None)
    max_end : ``pd.Timestamp``, optional, (default=None)

    Returns
    -------
    bool
    """
    min_start = min_start if min_start is not None else pd.Timestamp("1980-01-01")
    max_end = max_end if max_end is not None else pd.Timestamp("2020-01-01")
    return _range_args_are_correct(
        date_range_args, min_length, max_length, min_start, max_end
    )


def compute_date_range(date_range_args: IndexRangeArgs) -> pd.DatetimeIndex:
    """ Returns a Pandas DatetimeIndex from the given ``date_range_args``.

    Parameters
    ----------
    date_range_args: ``IndexRangeArgs``, required

    Returns
    -------
    pd.DatetimeIndex
    """
    return pd.date_range(**date_range_args)


@st.composite
def timedelta_range_args_from(
    draw,
    start: pd.Timedelta,
    end: pd.Timedelta,
    min_length: int,
    max_length: int,
    arg_to_remove: Optional[str] = None,
) -> IndexRangeArgs:
    """ Strategy that computes arguments to call pd.timedelta_range() from
    start, end and max length of the resulting period
    It returns a dictionary with 3 out of the 3 keys 'start', 'end', 'periods'
    and 'freq'. One of them is automatically removed since pd.timedelta_range()
    accepts only 3 of the 4 inputs.

    Parameters
    ----------
    draw
    start : datetime, required
    end : datetime, required
    min_length : int, required
    max_length : int, required
    arg_to_remove: str, optional, (default=None)

    Returns
    -------
    IndexRangeArgs
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.
    """
    start, end = draw(pair_of_ordered_timedeltas(start, end))
    periods = draw(st.integers(min_value=min_length, max_value=max_length))
    freq = draw(available_freqs())
    element_to_exclude = (
        draw(samples_from(pandas_range_params))
        if arg_to_remove is None
        else arg_to_remove
    )

    args = {"start": start, "end": end, "periods": periods, "freq": freq}
    del args[element_to_exclude]

    return args


def timedelta_range_args_are_correct(
    timedelta_range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: pd.Timedelta = None,
    max_end: pd.Timedelta = None,
) -> bool:
    """ Returns True if the timedelta range arguments are correct.
    It checks 3 things:

    - if 'periods' is not a key it checks that the length of the resulting
    period index is below ``max_length``.
    - if 'start' is not a keyword it checks that the expected start timedelta
    is after ``min_start``
    - if 'end' is not a keyword it checks that the expected end timedelta is
    before ``max_end``

    Parameters
    ----------
    timedelta_range_args : ``IndexRangeArgs``, required
        dictionary with 3 out of the 4 keys 'start', 'end', 'periods' and
        'freq'.

    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``pd.Timestamp``, optional, (default=None)
    max_end : ``pd.Timestamp``, optional, (default=None)

    Returns
    -------
    bool
    """
    min_start = min_start if min_start is not None else pd.Timedelta(0)
    max_end = max_end if max_end is not None else pd.Timedelta("40Y")
    return _range_args_are_correct(
        timedelta_range_args, min_length, max_length, min_start, max_end
    )


def compute_timedelta_range(timedelta_range_args: IndexRangeArgs):
    """ Returns a Pandas TimedeltaIndex from the given 
    ``timedelta_range_args``.

    Parameters
    ----------
    timedelta_range_args: ``IndexRangeArgs``, required

    Returns
    -------
    pd.DatetimeIndex
    """
    return pd.timedelta_range(**timedelta_range_args)


def _range_args_are_correct(
    range_args: IndexRangeArgs,
    min_length: int = 0,
    max_length: int = 1000,
    min_start: Union[pd.Timedelta, datetime] = None,
    max_end: Union[pd.Timedelta, datetime] = None,
) -> bool:
    """Checks if range args are correct.

    Parameters
    ----------
    range_args: ``IndexRangeArgs``, required
    min_length : ``int``, optional, (default=0)
    max_length : ``int``, optional, (default=1000)
    min_start : ``Union[pd.Timedelta, datetime]``, optional, (default=None)
    max_end : ``Union[pd.Timedelta, datetime]``, optional, (default=None)

    Returns
    -------
    bool
    """
    try:
        if "periods" not in range_args:
            return min_length < expected_index_length_from(**range_args) < max_length
        elif "start" not in range_args:
            return expected_start_date_from(**range_args) >= min_start
        elif "end" not in range_args:
            return expected_end_date_from(**range_args) <= max_end
        else:
            return True
    except OverflowError:
        return False



"""
The :mod:`gtime.metrics` module contains a collection of different metrics.
"""

from .metrics import (smape, max_error, mse, log_mse, 
                      r_square, mae, mape, rmse, rmsle, gmae)

__all__ = [
    "smape",
    "max_error",
    "mse",
    "rmse",
    "log_mse",
    "rmsle",
    "r_square",
    "mae",
    "mape",
    "gmae",
]



from random import random

import numpy as np
import pandas as pd
from hypothesis import given, settings
from hypothesis.extra.numpy import arrays
from hypothesis.strategies import floats

from gtime.regressors import LinearRegressor


class TestLinearRegressor:
    def test_linear_regressor(self):
        train, test = train_test_dataframe()

        predictions = compute_predictions_for_train_test(train, test)
        expected = compute_expectation_from_test(test)

        np.testing.assert_array_almost_equal(predictions, expected, decimal=2)

    @settings(deadline=None)
    @given(
        arrays(
            dtype=float,
            shape=(100, 1),
            elements=floats(allow_nan=False, allow_infinity=False, width=16),
        )
    )
    def test_linear_regressor_random_array(self, random_array):
        train, test = train_test_dataframe(random_array)

        predictions = compute_predictions_for_train_test(train, test)
        expected = compute_expectation_from_test(test)

        np.testing.assert_array_almost_equal(predictions, expected, decimal=0)


def train_test_dataframe(
    random_array: np.ndarray = None,
) -> (pd.DataFrame, pd.DataFrame):
    random_array = (
        random_array if random_array is not None else [random() for _ in range(100)]
    )

    a1, a2, b = random() * 10, random() * 100, 2 * (1 - random())

    df = pd.DataFrame()
    df["x1"] = list(range(100))
    df["x2"] = random_array
    df["y"] = [b + a1 * t for t in range(100)]
    df["y"] = df["y"] + a2 * df["x2"]

    train = df[:90]
    test = df[90:]

    return train, test


def compute_predictions_for_train_test(
    train: pd.DataFrame, test: pd.DataFrame
) -> np.ndarray:
    lr = LinearRegressor()

    lr.fit(train[["x1", "x2"]], train["y"], x0=[0, 0, 0])

    preds_y = lr.predict(test[["x1", "x2"]])
    preds_y = preds_y / np.sum(preds_y)

    return preds_y


def compute_expectation_from_test(test: pd.DataFrame) -> np.ndarray:
    test_y = test["y"].values
    test_y = test_y / np.sum(test_y)
    return test_y
